import io
import re
import shutil
import sys
from pathlib import Path
from typing import Any
from urllib.request import urlopen
from zipfile import ZipFile

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from tqdm import tqdm

_BASEURL = (
    'https://github.com/openslide/openslide-winbuild'
    '/releases/download/v{version}/openslide-win64-{version}.zip'
)
_FILTERS = {
    '20221217': r'(libjpeg-62|libtiff-6|zlib1).dll',
    '20231011': r'.*\.dll',
}
_URLS = {
    _BASEURL.format(version=version): re.compile(filter_)
    for version, filter_ in _FILTERS.items()
}
_TARGET = 'src/bipl/io/libs'


def _url_to_io(url: str) -> io.BytesIO:
    r = urlopen(url)

    total = int(value) if (value := r.headers['content-length']) else None
    buf = io.BytesIO()
    with tqdm.wrapattr(r, 'read', total, desc='retrieve libraries') as fp:
        shutil.copyfileobj(fp, buf)
    return buf


def _download_dlls(folder: Path) -> None:
    """Download archive, extract DLLs and place them into target folder"""
    if sys.platform != 'win32' or folder.exists():
        return

    folder.mkdir(parents=True, exist_ok=True)
    try:
        for url, regex in _URLS.items():
            with ZipFile(_url_to_io(url)) as zf:
                for zfpath in zf.namelist():
                    if not regex.fullmatch(Path(zfpath).name):
                        continue
                    with (
                        zf.open(zfpath) as src,
                        (folder / Path(zfpath).name).open('wb') as dst,
                    ):
                        shutil.copyfileobj(src, dst)
    except BaseException:
        for p in folder.glob('*.dll'):
            p.unlink()
        folder.rmdir()
        raise


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        if self.target_name != 'wheel':
            return

        build_data |= {
            'tag': 'py3-none-win_amd64',  # lock platform
            'pure_python': False,  # got DLLs inside
        }
        _download_dlls(Path(self.root, _TARGET))



__all__ = ['Paged', 'fopen']

import asyncio
import atexit
import mmap
import os
from contextlib import AsyncExitStack
from io import BytesIO
from threading import Thread
from typing import Protocol

import httpx
from glow import call_once, memoize
from pydantic import HttpUrl, TypeAdapter, ValidationError

from bipl._env import env
from bipl._types import Span
from bipl.ops._util import merge_intervals

_URL_ADAPTER = TypeAdapter(HttpUrl)


class Paged(Protocol):
    def pread(self, n: int, offset: int, /) -> bytes: ...


def fopen(path: str) -> Paged:
    try:  # Process as URL
        return _RemoteIO(path)

    except ValidationError:  # Got local filename
        return _MappedIO(path)


class _MappedIO:
    __slots__ = ('m',)

    def __init__(self, path: str) -> None:
        fd = os.open(path, os.O_RDONLY)
        try:
            self.m = mmap.mmap(fd, 0, access=mmap.ACCESS_READ)
        finally:
            os.close(fd)

    def __hash__(self) -> int:  # Used by _Level.tile:shared_call
        return hash(self.m)

    def __eq__(self, rhs) -> bool:  # Used by _Level.tile:shared_call
        return type(self) is type(rhs) and self.m == rhs.m

    def pread(self, n: int, offset: int, /) -> bytes:
        data = self.m[offset : offset + n]
        return data


class _RemoteIO:
    __slots__ = ('_cached', '_client', '_loop', 'url')

    def __init__(self, url: str) -> None:
        _URL_ADAPTER.validate_python(url)
        self.url = httpx.URL(url)

        self._loop, self._client = _aget_client()
        asyncio.run_coroutine_threadsafe(self._peek(), self._loop).result()

        self._cached = memoize(
            capacity=env.BIPL_TIFF_NUM_BLOCKS, bytesize=False, batched=True
        )(self._get_many)

    def __hash__(self) -> int:  # Used by _Level.tile:shared_call
        return hash(self.url)

    def __eq__(self, rhs) -> bool:  # Used by _Level.tile:shared_call
        return type(self) is type(rhs) and self.url == rhs.url

    def pread(self, n: int, offset: int, /) -> bytes:
        blk_size = env.BIPL_TIFF_BLOCK_SIZE
        lo = offset // blk_size
        hi = (offset + n + blk_size - 1) // blk_size

        s = BytesIO()
        for blk in self._cached([(o, o + 1) for o in range(lo, hi)]):
            s.write(blk)

        s.seek(offset - lo * blk_size)
        return s.read(n)

    def _get_many(self, spans: list[Span]) -> list[bytes]:
        # Combine spans with common border
        pos, spans = merge_intervals(spans)

        # Get long chunks
        fp = BytesIO()
        asyncio.run_coroutine_threadsafe(
            self._update_all(fp, spans), self._loop
        ).result()

        # Split back to blocks and order
        fp.seek(0)
        rs = [b''] * len(pos)
        for i in pos:
            rs[i] = fp.read(env.BIPL_TIFF_BLOCK_SIZE)
        return rs

    async def _peek(self) -> None:
        # Ensure range ok
        r = await self._client.head(self.url, headers={'Accept': '*/*'})
        r.raise_for_status()
        if r.headers.get('Accept-Ranges') != 'bytes':
            raise ValueError(
                f'Remote "{self.url}" does not support Range header'
            )

    async def _update_all(self, s: BytesIO, spans: list[Span]) -> None:
        fs = (self._update(start, stop) for start, stop in spans)
        for blk in await asyncio.gather(*fs):
            s.write(blk)

    async def _update(self, start: int, stop: int) -> bytes:
        blk_size = env.BIPL_TIFF_BLOCK_SIZE
        start *= blk_size
        stop *= blk_size

        async with self._client.stream(
            'GET', self.url, headers={'Range': f'bytes={start}-{stop - 1}'}
        ) as r:
            r.raise_for_status()
            if r.status_code != 206:
                raise ValueError(
                    f'Unexpected status: {r.status_code} for {self.url}'
                )
            return await r.aread()


@call_once
def _aget_client() -> tuple[asyncio.AbstractEventLoop, httpx.AsyncClient]:
    loop = asyncio.new_event_loop()
    Thread(target=loop.run_forever, daemon=True).start()

    aes = AsyncExitStack()
    atexit.register(_afinalize, aes, loop)

    aclient = httpx.AsyncClient(http1=False, http2=True, follow_redirects=True)
    asyncio.run_coroutine_threadsafe(
        aes.enter_async_context(aclient), loop
    ).result()

    return loop, aclient


def _afinalize(aes: AsyncExitStack, loop: asyncio.AbstractEventLoop) -> None:
    asyncio.run_coroutine_threadsafe(aes.aclose(), loop).result()



__all__ = ['Gdal', 'url_to_gdal', 'url_to_safe_gdal']

import os
from dataclasses import dataclass
from threading import Lock
from urllib.parse import urlencode

import numpy as np
from osgeo import gdal, gdal_array
from pydantic import AnyHttpUrl, TypeAdapter, ValidationError

from bipl._types import Span

from ._slide_bases import Driver, ImageLevel, PartsMixin
from ._util import gdal_parse_mpp

gdal.SetConfigOption('AWS_VIRTUAL_HOSTING', 'FALSE')
gdal.SetConfigOption('CPL_VSIL_CURL_USE_HEAD', 'NO')  # use_head
gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'EMPTY_DIR')  # list_dir
gdal.SetConfigOption('GDAL_HTTP_UNSAFESSL', 'YES')
gdal.SetConfigOption('GDAL_HTTP_MAX_RETRY', '3')  # max_retry
gdal.UseExceptions()

_url_adapter = TypeAdapter(AnyHttpUrl)


def url_to_gdal(url: AnyHttpUrl | str, /, **query) -> str:
    if not query:
        return f'/vsicurl/{url}'
    return f'/vsicurl?{urlencode(query | {"url": str(url)})}'


def url_to_safe_gdal(url: AnyHttpUrl | str, /) -> str:
    return url_to_gdal(url, list_dir='no', use_head='no', max_retry='3')


def _fix_if_url(s: str, /) -> str:
    if os.path.isfile(s) or s.startswith('/vsicurl'):
        return s
    try:
        return url_to_safe_gdal(_url_adapter.validate_python(s))
    except ValidationError:
        raise FileNotFoundError(f'Neither a fs path nor URL: {s!r}') from None


# TODO: handle associated images via `base.Image`


@dataclass(frozen=True, kw_only=True)
class _Level(PartsMixin, ImageLevel):
    g: 'Gdal'
    bands: tuple[gdal.Band, ...]

    def part(self, *loc: Span) -> np.ndarray:
        box, valid_box, shape = self._unpack_2d_loc(*loc)

        (y0, y1), (x0, x1) = valid_box.tolist()
        if y0 == y1 or x0 == x1:  # Patch is outside slide
            return np.broadcast_to(self.g.bg_color, (*shape, 3))

        h, w = y1 - y0, x1 - x0
        c = self.g.num_channels
        chw = np.empty((c, h, w), 'u1')

        with self.g.lock:
            for b, hw in zip(self.bands, chw[:, None, :, :]):
                gdal_array.BandReadAsArray(b, x0, y0, w, h, buf_obj=hw)

        # TODO: add pad if necessary
        rgb = chw.transpose(1, 2, 0)
        rgb = self._expand(rgb, valid_box, box, self.g.bg_color)
        return self._postprocess(rgb)


class Gdal(Driver):
    def __init__(self, path: str) -> None:
        path = _fix_if_url(path)
        self.ds: gdal.Dataset = gdal.OpenEx(path, gdal.GA_ReadOnly)

        drv: gdal.Driver = self.ds.GetDriver()
        if drv.ShortName != 'GTiff':
            raise ValueError(f'Unsupported driver {drv.ShortName}')

        self.num_channels = self.ds.RasterCount
        if self.num_channels not in {3, 4}:
            raise ValueError('Unknown colorspace')

        self._bands: tuple[gdal.Band, ...] = tuple(
            self.ds.GetRasterBand(i + 1) for i in range(self.num_channels)
        )
        self.dtype = np.dtype(gdal_array.flip_code(self._bands[0].DataType))
        if self.dtype != 'u1':
            raise ValueError(f'Unsupported dtype: {self.dtype}')

        self.meta = self.ds.GetMetadata().copy()

        self.bg_color = np.full(3, 255, 'u1')  # TODO: parse tags
        self.lock = Lock()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({id(self.ds):0x})'

    def get_mpp(self) -> float | None:
        if mpp := gdal_parse_mpp(self.meta):
            return float(np.mean(mpp))
        return None

    def __len__(self) -> int:
        return 1 + self.ds.GetRasterBand(1).GetOverviewCount()

    def __getitem__(self, index: int) -> _Level:
        bands: tuple[gdal.Band, ...] = (
            self._bands
            if index == 0
            else tuple(b.GetOverview(index - 1) for b in self._bands)
        )
        shape = (bands[0].YSize, bands[0].XSize, self.num_channels)
        return _Level(shape, g=self, bands=bands)

    def keys(self) -> list[str]:
        return []  # TODO: fill if GDAL can detect auxilary images



__all__ = ['load_library']

import os
import sys
from ctypes import CDLL
from pathlib import Path

_root = (Path(__file__).parent / 'libs').as_posix()


def _load_library(name: str, version: int) -> CDLL:
    if sys.platform != 'win32':
        return CDLL(f'{name}.so.{version}')

    with os.add_dll_directory(_root):
        return CDLL(f'{_root}/{name}-{version}.dll')


def load_library(name: str, *versions: int) -> CDLL:
    errors = []
    for v in versions:
        try:
            return _load_library(name, v)
        except (OSError, FileNotFoundError) as exc:
            errors.append(exc)
    raise errors[-1]



__all__ = ['BlendCropper', 'Decimator', 'Reconstructor', 'Stripper', 'Zipper']

from collections import defaultdict, deque
from collections.abc import Callable, Generator, Iterable, Iterator
from dataclasses import dataclass
from functools import partial

import numpy as np
import numpy.typing as npt

from bipl._types import NDIndex, NumpyLike, Shape, Tile, Vec

from ._util import crop_to, rescale_crop

_Op = Callable[[Tile], list[Tile]]


@dataclass(frozen=True, slots=True)
class Stripper:
    """Strips regions outside of `data.shape`. Produces non-square patches"""

    shape: Shape

    def __call__(self, tile: Tile) -> Tile:
        idx, vec, data = tile
        vec, data = crop_to(vec, data, self.shape)
        return Tile(idx=idx, vec=vec, data=data)


@dataclass(frozen=True, slots=True)
class Decimator:
    stride: int

    def __call__(self, tile: Tile) -> Tile:
        stride = self.stride
        vec = tuple(c // stride for c in tile.vec)
        data = tile.data[::stride, ::stride]
        return Tile(idx=tile.idx, vec=vec, data=data)


@dataclass(frozen=True, slots=True)
class Zipper:
    v: NumpyLike
    v_scale: float
    interpolation: int

    def __call__(
        self,
        tile: Tile,
    ) -> tuple[NDIndex, Vec, np.ndarray, np.ndarray]:
        tsize = tile.data.shape[:2]
        if not tile.data.size:
            _, _, *extra_dims = self.v.shape
            empty = np.empty((*tsize, *extra_dims), self.v.dtype)
            return tile.idx, tile.vec, tile.data, empty

        loc = ((c0, c0 + size) for c0, size in zip(tile.vec, tsize))
        r = rescale_crop(
            self.v, *loc, scale=self.v_scale, interpolation=self.interpolation
        )
        return tile.idx, tile.vec, tile.data, r


class Reconstructor:
    """Rebuilds full tiles from parts"""

    def __init__(self, overlap: int, cells: npt.NDArray[np.bool_]):
        assert overlap
        self.overlap = overlap
        self.cells = np.pad(cells, (0, 1))
        self.row = defaultdict[int, np.ndarray]()

    def __call__(self, t: Tile) -> Tile:
        (iy, ix), (y, x), part = t
        ov = self.overlap

        # Lazy init, first part is always whole
        if self.row.default_factory is None:
            self.row.default_factory = partial(
                np.empty, part.shape, part.dtype
            )

        if (tile := self.row.pop(ix, None)) is not None:
            # Incoming tile is always bottom-right aligned to the full one
            dy, dx = (tile.shape[a] - part.shape[a] for a in (0, 1))
            tile[dy:, dx:] = part
            y -= dy
            x -= dx
        else:
            tile = part

        # Use current right edge as new left edge
        if self.cells[iy, ix + 1]:
            self.row[ix + 1][:, :ov] = tile[:, -ov:]

        # Use current bottom edge as new top edge
        if self.cells[iy + 1, ix]:
            if ix and self.cells[iy + 1, ix - 1]:  # Don't write angle twice
                self.row[ix][:ov, ov:] = tile[-ov:, ov:]
            else:
                self.row[ix][:ov] = tile[-ov:]

        return Tile(idx=(iy, ix), vec=(y, x), data=tile)


class BlendCropper:
    """
    Blends edges/angles of overlapping tiles.
    Returns non-overlapping parts of tiles.

    All parts form non-uniform rectangular grid
    (i.e. all parts with same `iy` have same `shape[0]` and same `y`).

    NOTE: Each tile can make up to 4 parts.
    """

    __slots__ = ('call', 'cells', 'overlap', 'size')

    size: int
    cells: dict[NDIndex, npt.NDArray[np.bool_]]

    def __init__(self, cells: npt.NDArray[np.bool_], overlap: int, step: int):
        if overlap > step:
            raise ValueError('Tile overlap must be less or equal to tile step')
        self.overlap = overlap

        ih, iw = cells.shape
        # -> (2h+1 2w+1)
        cells = cells.repeat(2, 0).repeat(2, 1)
        cells = np.pad(cells, (0, 1))
        # -> (2h 2w)
        cells = np.lib.stride_tricks.sliding_window_view(cells, (2, 2))
        cells = np.bitwise_and.reduce(cells, axis=(-2, -1))
        # -> (h w 2 2) [[this, next], [next row, next diag]]
        cells = cells.reshape(ih, 2, iw, 2).transpose(0, 2, 1, 3)

        c00 = cells[:, :, 0, 0]
        sizes = (c00[:, :, None, None] & cells).sum((0, 1))

        if overlap < step:
            self.call = self._3x3
            self.size = sizes.sum()
        else:
            self.call = self._2x2
            self.size = sizes[1, 1]

        # (iy, ix) -> 2x2 of bool
        iys, ixs = c00.nonzero()
        self.cells = {
            (iy, ix): m
            for iy, ix, m in zip(iys.tolist(), ixs.tolist(), cells[iys, ixs])
        }

    def __call__(self, src: Iterable[Tile]) -> Iterator[Tile]:
        """Merge and split tiles."""
        # TODO: relax input/output order
        return self.call(src)

    def _2x2(self, src: Iterable[Tile]) -> Iterator[Tile]:
        ops = defaultdict[NDIndex, list[_Op]](list)
        ov = self.overlap

        for tile in src:  # NOTE: supports C and F-ordering
            iy, ix = idx = tile.idx

            for op in ops.pop(idx, []):  # full top & left (3)
                yield from op(tile)

            # TODO: bag ops
            if self.cells[iy, ix][1, 1]:  # bottom-right (1)
                u = _angle(ov, idx, tile.data)
                next(u)
                ops[iy, ix + 1].append(u.send)
                ops[iy + 1, ix].append(u.send)
                ops[iy + 1, ix + 1].append(partial(_final_send, u))

    def _3x3(self, src: Iterable[Tile]) -> Iterator[Tile]:
        last_iy = -1
        buf = deque[Tile]()
        ops = defaultdict[NDIndex, list[_Op]](list)
        ov = self.overlap

        for tile in src:  # NOTE: only C-ordering is supported.
            iy, ix = idx = tile.idx

            if iy != last_iy:  # Start/end row
                while buf:
                    yield buf.popleft()
                last_iy = iy  # NOTE: this one locks iteration order

            for op in ops.pop(idx, []):  # full top & left (5)
                yield from op(tile)
            (_, c01), (c10, c11) = self.cells[iy, ix].tolist()

            buf.append(_center(ov, tile))  # center (1)

            if c01:  # right (1)
                ops[iy, ix + 1].append(partial(_vert_edge, ov, buf, tile.data))

            if c10:  # bottom (1)
                ops[iy + 1, ix].append(partial(_horz_edge, ov, tile.data))

            if c11:  # bottom-right (1)
                u = _angle(ov, (iy * 2 + 1, ix * 2 + 1), tile.data)
                next(u)
                ops[iy, ix + 1].append(u.send)
                ops[iy + 1, ix].append(u.send)
                ops[iy + 1, ix + 1].append(partial(_final_send, u))

        while buf:
            yield buf.popleft()  # End row


def _center(pad: int, tile: Tile) -> Tile:
    # [- - -]
    # [- R -]
    # [- - -]
    (ix, iy), (y, x), data = tile
    return Tile((2 * ix, 2 * iy), (y + pad, x + pad), data[pad:-pad, pad:-pad])


def _vert_edge(
    pad: int, buf: deque[Tile], left: np.ndarray, tile: Tile
) -> list[Tile]:
    # [- - -]    [- - -]
    # [- - X] -> [R - -]
    # [- - -]    [- - -]
    (iy, ix), (y, x), right = tile
    tile = Tile(
        (iy * 2, ix * 2 - 1),
        (y + pad, x),
        left[pad:-pad, -pad:] + right[pad:-pad, :pad],
    )
    buf.append(tile)
    return []


def _horz_edge(pad: int, top: np.ndarray, tile: Tile) -> list[Tile]:
    # [- - -]    [- R -]
    # [- - -] -> [- - -]
    # [- X -]    [- - -]
    (iy, ix), (y, x), bottom = tile
    tile = Tile(
        (2 * iy - 1, 2 * ix),
        (y, x + pad),
        top[-pad:, pad:-pad] + bottom[:pad, pad:-pad],
    )
    return [tile]


def _angle(
    pad: int, idx: NDIndex, data: np.ndarray
) -> Generator[list[Tile], Tile, list[Tile]]:
    # [- . -]    [- . -]    [- . X]    [R . -]
    # [. . .] -> [. . .] -> [. . .] -> [. . .]
    # [- . X]    [X . -]    [- . -]    [- . -]

    # top-left + top-right
    data = data[-pad:, -pad:] + (yield []).data[-pad:, :pad]
    # bottom-left
    data += (yield []).data[:pad, -pad:]
    # bottom-right
    _, vec, last = yield []
    data += last[:pad, :pad]

    return [Tile(idx, vec, data)]


def _final_send(
    coro: Generator[list[Tile], Tile, list[Tile]], tile: Tile
) -> list[Tile]:
    try:
        coro.send(tile)
    except StopIteration as stop:
        return stop.value
    else:
        raise RuntimeError('Coroutine not stopped')



__all__ = ['Slide']

import importlib
import os
import warnings
from bisect import bisect_right
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import Literal, final, overload
from warnings import warn

import numpy as np
from glow import memoize, shared_call, weak_memoize

from bipl import cov, env
from bipl._types import Patch, Shape, Span, Vec
from bipl.ops import Normalizer, normalize_loc, rescale_crop

from ._slide_bases import Driver, Image, ImageLevel, PartMixin

# TODO: inside Slide.open import ._slide.registry,
# TODO: and in ._slide.registry do registration and DLL loading
# TODO: to make Slide export not require DLL presence


def _load_drivers() -> None:
    if os.name == 'nt':
        gdal_warn = (
            'Ensure it was properly installed or acquire it manually from '
            'https://github.com/cgohlke/geospatial-wheels/releases'
        )
    else:
        gdal_warn = (
            'Ensure it was properly installed, and run '
            '"pip install wheel && pip install gdal==`gdal-config --version`"'
        )

    imports = {
        'openslide': ('._openslide', 'Openslide', ''),
        'tiff': ('._tiff', 'Tiff', ''),
        'gdal': ('._gdal', 'Gdal', gdal_warn),
    }
    drivers: dict[str, type[Driver]] = {}
    for drvname in [*env.BIPL_DRIVERS]:
        # Find
        if (imp := imports.get(drvname)) is None:
            raise ValueError(f'Unknown driver: {drvname}')

        # Import
        modname, attrname, extramsg = imp
        try:
            mod = importlib.import_module(modname, __package__)
        except ImportError as exc:
            msg = f'"{drvname}" driver failed to load ({exc}). {extramsg}'
            warn(msg, stacklevel=2)
            env.BIPL_DRIVERS.discard(drvname)
        else:
            # Get driver
            drivers[drvname] = getattr(mod, attrname)

    os.environ['BIPL_DRIVERS'] = str([*env.BIPL_DRIVERS]).replace("'", '"')
    if not drivers:
        raise ImportError('No drivers loaded')

    for drvname, pat in [  # LIFO, last driver takes priority
        ('gdal', r'^.*\.tiff?$'),
        ('openslide', r'^.*\.(bif|mrxs|ndpi|scn|svs(|slide)|tiff?|vms|vmu)$'),
        ('tiff', r'^.*\.(bif|ndpi|svs|tiff?)$'),
        ('gdal', r'^(/vsicurl|(https?|ftp)://).*$'),
    ]:
        if drv := drivers.get(drvname):
            drv.register(pat)


_load_drivers()


# Merge duplicate calls
@shared_call
# Reuse result if it's already exist, but used by someone else
@weak_memoize
# Keep LRU for unused results
@memoize(capacity=env.BIPL_CACHE, policy='lru', bytesize=False)
def _cached_open(path: str) -> 'Slide':
    if tps := Driver.find(path):
        errors: list[Exception] = []
        for tp in reversed(tps):  # Loop over types to find non-failing
            try:
                return Slide.from_file(path, tp)
            except (ValueError, TypeError) as exc:
                errors.append(exc)
        raise ExceptionGroup('Cannot open file', errors) from None

    raise ValueError(f'Unknown file format {path}')


@final
@dataclass(frozen=True)
class Slide(PartMixin):
    """Usage:
    ```
    slide = Slide.open('test.svs')
    shape: tuple[int, ...] = slide.shape
    downsamples: tuple[int, ...] = slide.downsamples

    # Get numpy.ndarray
    image: np.ndarray = slide[:2048, :2048]
    ```
    """

    # TODO: check if memory leak
    # TODO: add __enter__/__exit__/close to make memory management explicit
    # TODO: call .close in finalizer
    path: str
    shape: Shape
    downsamples: tuple[int, ...]
    mpp: float | None
    driver: str
    bbox: tuple[slice, ...] = field(repr=False)
    levels: Sequence[ImageLevel] = field(repr=False)
    extras: Mapping[str, Image] = field(repr=False)
    dtype = np.dtype(np.uint8)

    @classmethod
    def from_file(
        cls,
        path: str,
        driver_fn: Callable[[str], Driver],
        mpp: float | None = None,
    ) -> 'Slide':
        driver = driver_fn(path)

        num_images = len(driver)
        if num_images == 0:
            raise ValueError('Empty file')

        # Retrieve all sub-images
        levels: list[ImageLevel] = []
        extras: dict[str, Image] = {}
        for idx in range(num_images):
            match driver[idx]:
                case ImageLevel() as im:
                    levels.append(im)
                case Image(key=str(key)) as im:
                    extras[key] = im
                case _:
                    continue
        extras |= driver.named_items()

        downsamples, levels = driver.build_pyramid(levels)

        if mpp is None:  # If no override is passed, use native if present
            mpp = driver.get_mpp()

        if mpp is not None and env.BIPL_MPP_Q:
            mpp = 2 ** (round(np.log2(mpp) * env.BIPL_MPP_Q) / env.BIPL_MPP_Q)

        return Slide(
            path=path,
            shape=levels[0].shape,
            mpp=mpp,
            downsamples=downsamples,
            driver=type(driver).__name__,
            bbox=driver.bbox,
            levels=levels,
            extras=extras,
        )

    def icc(self) -> 'Slide':
        changes = {}
        if icc_0 := self.levels[0].icc:
            changes['levels'] = tuple(il.apply(icc_0) for il in self.levels)
        if any(e.icc for e in self.extras.values()):
            changes['extras'] = {
                t: e.apply(e.icc) if e.icc else e
                for t, e in self.extras.items()
            }
        return replace(self, **changes) if changes else self

    def norm(
        self,
        mpp: float = 64,
        weight: float = 1.0,
        channels: Literal['L', 'Lab', 'ab'] = 'ab',
        r: float = 0.5,
    ) -> 'Slide':
        ref = self.resample(mpp=mpp).numpy()
        normalize = Normalizer(ref, weight=weight, channels=channels, r=r)

        levels = tuple(il.apply(normalize) for il in self.levels)
        extras = {
            k: i.apply(normalize) if k == 'thumbnail' else i
            for k, i in self.extras.items()
        }
        return replace(self, levels=levels, extras=extras)

    def mpp_or_error(self) -> float:
        if self.mpp is None:
            raise ValueError('Slide`s MPP is unknown')
        return self.mpp

    @property
    def pools(self) -> Sequence[int]:
        warnings.warn(
            '"Slide.pools" is deprecated. Use "Slide.downsamples"',
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.downsamples

    @property
    def spacing(self) -> float | None:
        warnings.warn(
            '"Slide.spacing" is deprecated. Use "Slide.mpp"',
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.mpp

    def __reduce__(self) -> tuple:
        return Slide.open, (self.path,)

    def best_level_for(
        self,
        downsample: float,
        /,
        *,
        tol: float = 0.01,
    ) -> tuple[int, ImageLevel]:
        """Gives the most detailed LOD below `downsample`"""
        k = max(1, 1 + tol)
        idx = max(bisect_right(self.downsamples, downsample * k) - 1, 0)
        ds = self.downsamples[idx]
        lv = self.levels[idx]

        # Make 2^k level as close as possible to target
        return lv.decimate(downsample, ds)

    def resample(self, mpp: float, *, tol: float = 0.01) -> ImageLevel:
        """Resample slide to specific resolution"""
        downsample = mpp / self.mpp_or_error()
        return self.pool(downsample, tol=tol)

    def pool(self, downsample: float, /, *, tol: float = 0.01) -> ImageLevel:
        """Use like `slide.pool(4)[y0:y1, x0:x1]` call"""
        ds, lvl = self.best_level_for(downsample, tol=tol)
        if ds == downsample:
            return lvl
        return lvl.rescale(ds / downsample)

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Retrieve image patch from maximum resolution"""
        y_loc, x_loc, (c_lo, c_hi) = normalize_loc(key, self.shape)
        return self.part(y_loc, x_loc)[:, :, c_lo:c_hi]

    def parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[Patch]:
        return self.levels[0].parts(locs, max_workers=max_workers)

    def numpy(self) -> np.ndarray:
        return self.levels[0].numpy()

    @overload
    def at(
        self,
        z0_yx_offset: Vec,
        dsize: int | Shape,
        *,
        scale: float,
        tol: float = ...,
    ) -> np.ndarray: ...

    @overload
    def at(
        self,
        z0_yx_offset: Vec,
        dsize: int | Shape,
        *,
        mpp: float,
        tol: float = ...,
    ) -> np.ndarray: ...

    def at(
        self,
        z0_yx_offset: Vec,
        dsize: int | Shape,
        *,
        scale: float | None = None,
        mpp: float | None = None,
        tol: float = 0.01,
    ) -> np.ndarray:
        """Read square region starting with offset"""
        dsize = dsize if isinstance(dsize, Sequence) else (dsize, dsize)
        if len(dsize) != 2:
            raise ValueError(f'dsize should be 2-tuple or int. Got {dsize}')

        if scale is None:
            if mpp is None:
                raise ValueError('Only one of zoom/mpp should be None')
            scale = self.mpp_or_error() / mpp

        cov.update(self.path, z0_yx_offset, dsize, scale)
        ds, lvl = self.best_level_for(1 / scale, tol=tol)

        yx_offset = tuple(int(c * scale) for c in z0_yx_offset)
        loc = tuple((c, c + size) for c, size in zip(yx_offset, dsize))
        return rescale_crop(lvl, *loc, scale=1 / ds / scale, interpolation=1)

    def extra(self, name: str) -> np.ndarray | None:
        if im := self.extras.get(name):
            return im.numpy()
        return None

    def thumbnail(self) -> np.ndarray:
        return self.extras.get('thumbnail', self.levels[-1]).numpy()

    @classmethod
    def open(
        cls,
        anypath: Path | str,
        /,
        *,
        icc: bool = env.BIPL_ICC,
        norm: bool | float = env.BIPL_NORM,
    ) -> 'Slide':
        """Open multi-scale image."""
        if isinstance(anypath, Path):  # Filesystem
            anypath = anypath.resolve().absolute().as_posix()
        s = _cached_open(anypath)
        if icc:
            s = s.icc()
        if not norm:
            return s
        return s.norm() if norm is True else s.norm(mpp=norm)



from importlib import import_module
from typing import TYPE_CHECKING

from ._dzi import Dzi

_exports = {
    '._slide': ['Slide'],
}
_submodule_by_name = {
    name: modname for modname, names in _exports.items() for name in names
}

if TYPE_CHECKING:
    from ._slide import Slide
else:

    def __getattr__(name: str):
        if modname := _submodule_by_name.get(name):
            mod = import_module(modname, __package__)
            globals()[name] = obj = getattr(mod, name)
            return obj
        raise AttributeError(f'No attribute {name}')

    def __dir__():
        return __all__


__all__ = ['Dzi', 'Slide']



__all__ = [
    'HasPartsAbc',
    'NDIndex',
    'NumpyLike',
    'Patch',
    'Shape',
    'Span',
    'Tile',
    'Vec',
]

from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from typing import NamedTuple, Protocol

import numpy as np

type NDIndex = tuple[int, ...]  # Index vector
type Shape = tuple[int, ...]  # N-dim shape
type Span = tuple[int, int]  # start/stop for `slice()`
type Vec = tuple[int, ...]  # N-dim radius-vector to some point


class NumpyLike(Protocol):
    @property
    def shape(self) -> Sequence[int]: ...

    @property
    def dtype(self) -> np.dtype: ...

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray: ...


class Tile(NamedTuple):
    idx: NDIndex
    vec: Vec
    data: np.ndarray

    def shape(self) -> Shape:
        return self.data.shape


class Patch(NamedTuple):
    loc: tuple[Span, ...]
    data: np.ndarray


class HasPartsAbc(ABC):
    @abstractmethod
    def parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[Patch]: ...



__all__ = ['Dzi']

import json
from typing import TYPE_CHECKING, Literal, NamedTuple

import cv2
import numpy as np

from bipl._types import NDIndex

if TYPE_CHECKING:
    from ._slide import Slide


class Dzi(NamedTuple):
    tile_size: int
    quality: int = 90
    fmt: Literal['jpeg', 'webp', 'avif'] = 'webp'

    def head(self, slide: 'Slide') -> str:
        h, w, _ = slide.shape
        mpp = slide.mpp
        image = {
            'xmlns': 'http://schemas.microsoft.com/deepzoom/2008',
            'Size': {
                'Height': h,
                'Width': w,
            },
            'MPP': mpp,
            'TileSize': self.tile_size,
            'Overlap': '0',
            'Format': self.fmt,
        }
        return json.dumps({'Image': image})

    def tile_at(
        self, slide: 'Slide', level: int, iy_ix: NDIndex
    ) -> np.ndarray:
        scale = 2 ** (level - max(slide.shape[:2]).bit_length())
        tile_0 = self.tile_size / scale
        offset_0 = tuple(ip * tile_0 for ip in iy_ix)
        return slide.at(offset_0, self.tile_size, scale=scale)

    def compress(self, rgb: np.ndarray) -> bytes:
        qtag = {
            'jpeg': cv2.IMWRITE_JPEG_QUALITY,
            'webp': cv2.IMWRITE_WEBP_QUALITY,
        }.get(self.fmt)
        if qtag is None:
            raise RuntimeError(f'Unknown format: {self.fmt}')

        bgr = cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)
        _, data = cv2.imencode(f'.{self.fmt}', bgr, [qtag, self.quality])
        return data.tobytes()



__all__ = ['Driver', 'Image', 'ImageLevel', 'PartMixin', 'PartsMixin']

import re
from abc import abstractmethod
from collections.abc import Callable, Iterator, Sequence
from dataclasses import dataclass, field, replace
from itertools import starmap
from typing import TYPE_CHECKING, Self, final

import cv2
import numpy as np
from glow import aceil, afloor, map_n

from bipl import env
from bipl._types import HasPartsAbc, Patch, Shape, Span, Tile
from bipl.ops import get_fusion, normalize_loc, resize

from ._util import round2

if TYPE_CHECKING:
    from ._util import Icc

_REGISTRY: dict[re.Pattern, list[type['Driver']]] = {}
_MIN_TILE = 256


@dataclass(frozen=True)
class Image:
    shape: Shape
    dtype = np.dtype(np.uint8)
    post: list[Callable[[np.ndarray], np.ndarray]] = field(
        default_factory=list
    )
    key: str | None = None

    def numpy(self) -> np.ndarray:
        """Convert to ndarray"""
        raise NotImplementedError

    @final
    def __array__(self) -> np.ndarray:
        """numpy.array compatibility"""
        return self.numpy()

    def apply(self, fn: Callable[[np.ndarray], np.ndarray]) -> Self:
        return replace(self, post=[*self.post, fn])

    def _postprocess(self, im: np.ndarray) -> np.ndarray:
        for fn in self.post:
            im = fn(im)
        return im

    @property
    def icc(self) -> 'Icc | None':
        return None


class PartMixin(HasPartsAbc):
    @final
    def part(self, *loc: Span) -> np.ndarray:
        [(_, a)] = self.parts([loc])
        return a


class PartsMixin(HasPartsAbc):
    @abstractmethod
    def part(self, *loc: Span) -> np.ndarray:
        """Reads crop of LOD. Overridable"""
        raise NotImplementedError

    @final
    def parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[Patch]:
        return map_n(
            lambda loc: Patch(loc, self.part(*loc)),
            locs,
            max_workers=max_workers,
        )


@dataclass(frozen=True)
class ImageLevel(Image, HasPartsAbc):
    key = None

    @final
    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Retrieve sub-image as array from set location"""
        y_loc, x_loc, (c_lo, c_hi) = normalize_loc(key, self.shape)
        [(_, a)] = self.parts([(y_loc, x_loc)])
        return a[:, :, c_lo:c_hi]

    @final
    def numpy(self) -> np.ndarray:
        """Retrieve whole image as array"""
        return self[:, :]

    def rescale(self, scale: float) -> 'ImageLevel':
        """
        Rescale image to set `scale`. Downscale if `scale` is less then 1,
        upscale otherwise.
        """
        if scale == 1:
            return self

        post = self.post
        if isinstance(self, ProxyLevel):  # Unwrap if already present
            scale = self.scale * scale
            base = self.base
        else:
            # Discard `post`
            base = self
            if prev := getattr(base, 'prev', None):
                base = replace(base, prev=replace(prev, post=[]))
            base = replace(base, post=[])

        h, w, c = base.shape
        h, w = (round(h * scale), round(w * scale))  # TODO: round/ceil/floor ?
        if scale <= 0.5:  # Downscale to more then 2x

            # NOTE: to use this datapath we must
            # - have 4^k downsamples (seen only in SVS)
            # - fail `octave()` call - always succedes unless SVS is choked
            #   (0s in tile sizes)
            downsample = 2 ** (int(1 / scale).bit_length() - 1)
            r_tile = max(_MIN_TILE // downsample, 1)
            bh, bw, bc = base.shape
            bh, bw = (
                (bh + downsample - 1) // downsample,
                (bw + downsample - 1) // downsample,
            )

            scale *= downsample
            base = TiledProxyLevel(
                (bh, bw, bc), base=base, downsample=downsample, r_tile=r_tile
            )

        if scale != 1 and base.shape != (h, w, c):
            return ProxyLevel((h, w, c), post=post, scale=scale, base=base)

        # Inject `post` back
        if prev := getattr(base, 'prev', None):
            base = replace(base, prev=replace(prev, post=post))
        return replace(base, post=post)

    def decimate(self, dst: float, src: int = 1) -> tuple[int, 'ImageLevel']:
        return (src, self)

    def _unpack_2d_loc(
        self, *loc: Span
    ) -> tuple[np.ndarray, np.ndarray, Shape]:
        box = np.array(loc)  # box[axis, {start, stop}]

        # Slices guarantied to be within image shape
        h, w = self.shape[:2]
        valid_box = box.clip(0, [[h], [w]])

        # Full output shape
        out_shape = tuple((box @ [-1, 1]).tolist())
        return box, valid_box, out_shape

    @staticmethod
    def _expand(
        rgb: np.ndarray,
        valid_box: np.ndarray,
        box: np.ndarray,
        bg_color: np.ndarray,
    ) -> np.ndarray:
        offsets = np.abs(valid_box - box)
        if offsets.any():
            top, bottom, left, right = offsets.ravel().tolist()
            rgb = cv2.copyMakeBorder(
                rgb,
                top,
                bottom,
                left,
                right,
                borderType=cv2.BORDER_CONSTANT,
                value=bg_color.tolist(),
            )
        return np.ascontiguousarray(rgb)


@dataclass(frozen=True, kw_only=True)
class ProxyLevel(PartMixin, ImageLevel):
    scale: float
    base: ImageLevel

    @property
    def interpolation(self) -> int:
        return cv2.INTER_LINEAR

    def parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[Patch]:
        scale = 1 / self.scale
        o_locs = np.asarray(locs, 'i4')  # (n yx lo/hi)
        n = o_locs.shape[0]

        i_locs_f = o_locs * scale
        i_locs = np.stack(  # (n yx lo/hi)
            [afloor(i_locs_f[:, :, 0], 'i4'), aceil(i_locs_f[:, :, 1], 'i4')],
            -1,
        )
        sizes = o_locs @ [-1, 1]  # (n yx)

        # Transformation matrix, (n xy xyc)
        mats = np.zeros((n, 2, 3), 'f4')
        mats[:, 0, 0] = scale
        mats[:, 1, 1] = scale
        mats[:, :, 2] = (
            i_locs_f[:, ::-1, 0] - i_locs[:, ::-1, 0] + (scale - 1) / 2
        )

        # Map input -> output
        i_locs_lst: list[tuple[Span, ...]]
        o_locs_lst: list[tuple[Span, ...]]
        i_locs_lst = [tuple(map(tuple, loc)) for loc in i_locs.tolist()]
        o_locs_lst = [tuple(map(tuple, loc)) for loc in o_locs.tolist()]
        i2o = dict(zip(i_locs_lst, zip(o_locs_lst, sizes.tolist(), mats)))

        kwargs = {
            'flags': self.interpolation | cv2.WARP_INVERSE_MAP,
            'borderMode': cv2.BORDER_CONSTANT,
        }

        # TODO: make this parallel
        def resample(iyx: tuple[Span, ...], i: np.ndarray) -> Patch:
            oyx, (h, w), mat = i2o[iyx]
            if not i.size:
                o = np.empty((h, w, *i.shape[2:]), i.dtype)
            else:
                o = cv2.warpAffine(i, mat, (w, h), **kwargs)  # type: ignore
            return Patch(oyx, self._postprocess(o))

        return starmap(resample, self.base.parts(i_locs_lst, max_workers))


@dataclass(frozen=True, kw_only=True)
class TiledProxyLevel(PartMixin, ImageLevel):
    base: ImageLevel
    downsample: int
    r_tile: int

    def parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[Patch]:
        # TODO: make `resize` & `postprocess` parallel
        for loc, a in zip(locs, self._parts(locs, max_workers)):
            yield Patch(loc, self._postprocess(a))

    def _parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[np.ndarray]:
        locs_a = np.asarray(locs, 'i4')  # (n yx lo/hi)
        s_locs = locs_a * self.downsample  # (n yx lo/hi)
        r_shapes = locs_a @ [-1, 1]  # (n yx)
        _, _, c = self.shape

        if not env.BIPL_TILED_POOLING:
            # Iterate over base, then resize
            pit = self.base.parts(s_locs.tolist(), max_workers)
            for p, r_shape in zip(pit, r_shapes.tolist()):
                # NOTE: resize is not concurrent
                yield resize(p.data, r_shape)
            return

        r_tile = self.r_tile
        s_tile = r_tile * self.downsample

        loc_ids: list[int] = []
        tgrids: list[np.ndarray] = []
        base_locs: list[tuple[Span, ...]] = []
        for i, s_loc in enumerate(s_locs):  # (yx lo/hi)
            ty, tx = (s_loc @ [-1, 1] + s_tile - 1) // s_tile
            tgrid = np.mgrid[:ty, :tx].reshape(2, -1).T  # (n yx)

            # (n yx lo/hi)
            st_locs = (
                (tgrid * s_tile + s_loc[..., 0])[..., None] + [0, s_tile]
            ).clip(
                *s_loc[None].T  # (yx lo/hi) -> lo(yx 1), hi(yx 1)
            )
            base_locs.extend(st_locs.tolist())
            loc_ids.extend(i for _ in tgrid)
            tgrids.append(tgrid)

        bufs: dict[int, list[np.ndarray]] = {i: [] for i, _ in enumerate(locs)}

        def build_part(i: int, buf: list[np.ndarray]) -> np.ndarray:
            r_shape = r_shapes[i]
            if buf:
                tgrid = tgrids[i]
                it = map(Tile, tgrid.tolist(), (tgrid * r_tile).tolist(), buf)
                # Stack tiles clipping out-of-shape regions
                image = get_fusion(it, r_shape)
                assert image is not None
            else:
                image = np.empty((*r_shape, c), self.dtype)
            return image

        # Iterate over base in tiles, resize, then stack via fusion
        # Should be effective for large downsamples of large levels (i.e. NDPI)
        prev = 0
        for i, (_, a) in zip(loc_ids, self.base.parts(base_locs, max_workers)):
            # NOTE: resize & fusion are not concurrent
            # Update buffer & yield done items on jumps
            bufs[i].append(resize(a, (r_tile, r_tile)))
            while prev != i:
                yield build_part(prev, bufs.pop(prev))
                prev += 1

        for i, buf in bufs.items():  # Finalize remaining
            yield build_part(i, buf)


class Driver:
    def get_mpp(self) -> float | None:
        raise NotImplementedError

    @final
    @classmethod
    def register(cls, regex: str) -> None:
        """Registers type builder for extensions. Last call takes precedence"""
        _REGISTRY.setdefault(re.compile(regex), []).append(cls)

    @final
    @staticmethod
    def find(path: str) -> list[type['Driver']]:
        tps: dict[type[Driver], None] = {}
        for pat, tps_ in _REGISTRY.items():
            if pat.match(path):
                for tp in tps_:
                    tps[tp] = None
        return [*tps]

    def __init__(self, path: str) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        """Count of indexed images, usually resolution images"""
        return 0

    def __getitem__(self, index: int) -> Image | None:
        """Gives indexed image"""
        raise NotImplementedError

    def named_items(self) -> dict[str, Image]:
        keys = self.keys()
        return {k: self.get(k) for k in keys}

    def keys(self) -> list[str]:
        """Names of associated images"""
        return []

    def get(self, key: str) -> Image:
        """Get assosiated image by key"""
        raise NotImplementedError

    @property
    def bbox(self) -> tuple[slice, ...]:
        return slice(None), slice(None)

    def build_pyramid(
        self, levels: Sequence[ImageLevel]
    ) -> tuple[tuple[int, ...], list[ImageLevel]]:
        if not levels:
            raise TypeError('No tiled layers present')

        full, *rest = levels
        pyramid = {1: full}
        for lv in rest:
            ds = round2(full.shape[0] / lv.shape[0])
            pyramid[ds] = lv

        downsamples = tuple(sorted(pyramid))
        levels = [pyramid[ds] for ds in downsamples]
        return downsamples, levels



__all__ = ['env']

from typing import Literal

from pydantic import ByteSize, HttpUrl, TypeAdapter
from pydantic_settings import BaseSettings

_bs_adapter = TypeAdapter(ByteSize)


class Env(BaseSettings):
    BIPL_DRIVERS: set[str] = {'gdal', 'tiff', 'openslide'}
    # Max slides opened
    BIPL_CACHE: int = 64
    # Max tiles cached per tiff slide
    BIPL_TILE_CACHE: ByteSize = _bs_adapter.validate_python('16 MiB')

    # `area`
    # - Downsample in single `cv2.resize(inter=AREA)` op.
    # - Uses `box` filter.
    # + No aliasing.
    # - Slow for large downsample factors.
    # box
    # - If downsample is more than 2X,
    #   do `cv2.resize(f=0.5)` till remaining downsample becomes less then 2X.
    # - Each step effectively filters image with [1 1] window,
    #   keeping 1/4 of pixels.
    # + Faster than `area` at least as 2x.
    # + Preserves alignment.
    # - Can lead to some aliasing.
    # gauss
    # - same as `box` but uses `cv2.pyrDown`
    # - Each step effectively filters image with [1 4 6 4 1] window,
    #   keeping 1/4 of pixels.
    # + No aliasing.
    # + Middle-ground in perf.
    # - Shifts image by 1/2 of pixel on each step.
    # TODO: implement [1 3 3 1] filtering
    BIPL_DOWN: Literal['area', 'box', 'gauss'] = 'gauss'  # Downsampling mode
    BIPL_SUBPIX: bool = True  # Subpixel resampling

    BIPL_TILED_POOLING: bool = False  # Do tiled pooling
    BIPL_ICC: bool = False  # Apply image color correction
    BIPL_NORM: bool = False  # Normalize luminance

    BIPL_MPP_Q: int = 12  # 2^(1/Q) MPP rounding (default 1.06x step)

    # Remote TIFF block size & cache size
    BIPL_TIFF_BLOCK_SIZE: ByteSize = _bs_adapter.validate_python('16 KiB')
    BIPL_TIFF_NUM_BLOCKS: int = 1000

    BIPL_COV_URL: HttpUrl | None = None


env = Env()



__all__ = [
    'floor2',
    'gdal_parse_mpp',
    'get_aperio_properties',
    'get_ventana_properties',
    'parse_xml',
    'round2',
    'unflatten',
]

from collections.abc import Mapping
from io import BytesIO
from itertools import zip_longest
from typing import Any

import cv2
import numpy as np
from lxml import etree
from lxml.etree import XMLParser, fromstring
from PIL.Image import fromarray
from PIL.ImageCms import buildTransform, createProfile

# ------------------------- flat list of properties --------------------------


def unflatten(d: Mapping[str, str]) -> dict[str, Any]:
    if not d:
        return {}

    tab = {ord('['): '.', ord(']'): None}
    pairs = sorted((_make_key(k, tab), v) for k, v in d.items())
    r: dict = {}

    for parts, v in pairs:
        rr = r
        for curr, next_ in zip_longest(parts, parts[1:], fillvalue=None):
            # Put list if next is list idx, otherwise put dict
            default: list | dict | str
            default = (
                v if next_ is None else ([] if isinstance(next_, int) else {})
            )

            if isinstance(curr, int):  # last was list idx
                if not isinstance(rr, list):  # Ambigous structure
                    break
                if curr >= len(rr):
                    rr.extend(None for _ in range(len(rr), curr + 1))
                    rr[curr] = default
                rr = rr[curr]
            else:  # last was dict key
                if not isinstance(rr, dict):  # Ambigous structure
                    break
                rr = rr.setdefault(curr, default)
    return r


def _make_key(s: str, tab: Mapping[int, str | None]) -> tuple[str | int, ...]:
    parts = s.translate(tab).split('.')
    return tuple(int(word) if word.isdigit() else word for word in parts)


# ---------------------------- aperio description ----------------------------


def get_aperio_properties(
    description: str, index: int = 0
) -> tuple[str, dict[str, str]] | None:
    if index == 0 and not description.startswith('Aperio'):
        return None
    header, *kv_pairs = description.split('|')

    header = '\n'.join(s.strip() for s in header.splitlines())

    meta = {}
    for i, kv in enumerate(kv_pairs):
        match kv.split('=', 1):
            case [str(k), str(v)]:
                meta[k.strip()] = v.strip()
            case ['']:
                continue
            case _:
                raise ValueError(
                    f'Cannot parse TIFF description line #{i}: '
                    f'{kv!r}, {description!r}'
                )

    return header, meta


# ----------------------------- xml description ------------------------------


def get_ventana_properties(s: bytes, index: int = 0) -> dict[str, str]:
    s = s.strip(b'\00')  # For safety
    t = fromstring(s, XMLParser(resolve_entities=False, no_network=True))
    if index == 0:
        if (root := t.find('iScan')) is not None:
            return dict(root.items())
        return {}
    return {'xmp': s.decode()}


def parse_xml(
    s: str,
    /,
    *,
    group: str = 'property',
    name: str = 'name',
    value: str = 'value',
) -> dict[str, str]:
    t = fromstring(s, XMLParser(resolve_entities=False, no_network=True))

    # Remove a namespace URI in the element's name
    for elem in t.getiterator():
        if not isinstance(elem, etree._Comment | etree._ProcessingInstruction):
            elem.tag = etree.QName(elem).localname

    # Remove unused namespace declarations
    etree.cleanup_namespaces(t, top_nsmap=None, keep_ns_prefixes=None)

    return {e.find(name).text: e.find(value).text for e in t.iter(group)}


# ----------------------------------------------------------------------------


def _gdal_parse_description(meta: Mapping[str, str]) -> dict[str, Any]:
    desc = meta.get('TIFFTAG_IMAGEDESCRIPTION')
    if not desc:
        return {}
    if get_aperio_properties(desc):
        raise ValueError('Aperio is not yet supported by GDAL driver')
    try:
        props = parse_xml(desc)
    except Exception:  # noqa: BLE001
        return {}
    return unflatten(props)


def gdal_parse_mpp(meta: Mapping) -> list[float]:
    props = _gdal_parse_description(meta)

    # TIFF tags
    res = [float(v) for a in 'XY' if (v := meta.get(f'TIFFTAG_{a}RESOLUTION'))]
    if res:
        match meta.get('TIFFTAG_RESOLUTIONUNIT'):
            case '2 (pixels/inch)':
                return [25_400 / r for r in res]
            case '3 (pixels/cm)':
                return [10_000 / r for r in res]

    # VIPS tags, always px/mm
    if res := [float(v) for tag in ('xres', 'yres') if (v := props.get(tag))]:
        return [1_000 / r for r in res]

    # Openslide tags
    if osd := props.get('openslide'):
        mpp = [float(v) for tag in ('mpp-x', 'mpp-y') if (v := osd.get(tag))]
        if mpp:
            return mpp

    # Aperio tags
    if mpp_ := props.get('MPP'):
        return [float(mpp_), float(mpp_)]

    return []


# ------------- contrast-limited adaptive histogram equalization -------------


def clahe(
    im: np.ndarray, cell_size: int = 64, clip_limit: float = 2.0
) -> np.ndarray:
    h, w = im.shape[:2]
    gh, gw = (h // cell_size) or 1, (w // cell_size) or 1
    cl = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=(gw, gh))

    ls, a, b = cv2.split(cv2.cvtColor(im, cv2.COLOR_RGB2LAB))
    ls = cl.apply(ls)
    return cv2.cvtColor(cv2.merge([ls, a, b]), cv2.COLOR_LAB2RGB)


# -------------------------- image color correction --------------------------


class Icc:
    def __init__(self, icc: bytes) -> None:
        f = BytesIO(icc)
        srgb = createProfile('sRGB')
        self._tf = buildTransform(f, srgb, inMode='RGB', outMode='RGB')

    def __call__(self, image: np.ndarray) -> np.ndarray:
        pil = fromarray(image)
        pil = self._tf.apply(pil)
        return np.array(pil, copy=False)


# -------------------------------- etc---------------------------------------


def round2(x: float) -> int:
    """Round to `2^k`, `k` is non-negative"""
    return floor2(x * _sqrt2)


def floor2(x: float) -> int:
    """Floor to `2^k`, `k` is non-negative"""
    x = max(x, 1)
    power = int(x).bit_length() - 1
    return 1 << power


_sqrt2 = 2**0.5



"""
Driver based on OpenSlide
- slow
- thread-safe
- compatible with formats: tiff/tif/svs, ndpi/vms/vmu, scn, mrxs, svsslide, bif
- relies on tile caching

To adjust tile cache size use BIPL_TILE_CACHE environment variable.
"""

# TODO: handle viewport offsets

__all__ = ['Openslide']

import atexit
import weakref
from collections.abc import Iterator
from ctypes import (
    POINTER,
    addressof,
    byref,
    c_char_p,
    c_double,
    c_int32,
    c_int64,
    c_ubyte,
    c_uint64,
    c_void_p,
)
from dataclasses import dataclass
from functools import cached_property

import cv2
import numpy as np
from packaging.version import Version

from bipl._env import env
from bipl._types import Span

from ._libs import load_library
from ._slide_bases import Driver, Image, ImageLevel, PartsMixin
from ._util import Icc, round2, unflatten

OSD = load_library('libopenslide', 1, 0)

sptr = POINTER(c_ubyte)
OSD.openslide_open.restype = sptr
OSD.openslide_close.argtypes = [sptr]

OSD.openslide_get_version.restype = c_char_p
OSD.openslide_get_error.restype = c_char_p
OSD.openslide_get_property_value.restype = c_char_p

OSD.openslide_get_property_names.restype = POINTER(c_char_p)
OSD.openslide_get_associated_image_names.restype = POINTER(c_char_p)

OSD.openslide_get_level_dimensions.argtypes = [
    sptr,  # openslide_t
    c_int32,  # level
    POINTER(c_int64),  # width
    POINTER(c_int64),  # height
]

OSD.openslide_get_level_downsample.argtypes = [sptr, c_int32]
OSD.openslide_get_level_downsample.restype = c_double
OSD.openslide_read_associated_image.argtypes = [sptr, c_char_p, c_void_p]

OSD.openslide_read_region.argtypes = [
    sptr,  # openslide_t
    c_void_p,  # buffer
    c_int64,  # x0
    c_int64,  # y1
    c_int32,  # level
    c_int64,  # width
    c_int64,  # height
]

_VERSION = None
_OSD_4X = False
if _version_bytes := OSD.openslide_get_version():
    _VERSION = Version(_version_bytes.decode())
    _OSD_4X = _VERSION >= Version('4.0')  # noqa: SIM300


def _init_4x_icc():
    if not _OSD_4X:
        return
    OSD.openslide_get_icc_profile_size.argtypes = [sptr]
    OSD.openslide_get_icc_profile_size.restype = c_int64
    OSD.openslide_read_icc_profile.argtypes = [sptr, c_void_p]

    OSD.openslide_get_associated_image_icc_profile_size.argtypes = [
        sptr,  # openslide_t
        c_char_p,  # name
    ]
    OSD.openslide_get_associated_image_icc_profile_size.restype = c_int64
    OSD.openslide_read_associated_image_icc_profile.argtypes = [
        sptr,  # openslide_t
        c_char_p,  # name
        c_void_p,  # buffer
    ]


def _init_4x_cache():
    if not _OSD_4X:
        return None
    OSD.openslide_cache_create.argtypes = [c_uint64]
    OSD.openslide_cache_create.restype = sptr
    OSD.openslide_cache_release.argtypes = [sptr]
    OSD.openslide_set_cache.argtypes = [sptr, sptr]

    cache = OSD.openslide_cache_create(env.BIPL_TILE_CACHE)
    atexit.register(OSD.openslide_cache_release, cache)
    return cache


_init_4x_icc()
_CACHE = _init_4x_cache()


def _ntas_to_iter(null_terminated_array_of_strings) -> Iterator[bytes]:
    if not null_terminated_array_of_strings:
        return
    i = 0
    while s := null_terminated_array_of_strings[i]:
        yield s
        i += 1


_LUT_CA = (np.outer(np.arange(256), np.arange(256)) // 255).astype('u1')


def _mbgra_to_rgb(bgra: np.ndarray, rgb_base: np.ndarray) -> np.ndarray:
    """
    Pre-multiplied BGRA to RGB. Alpha blending done with background.
    I.e. `RGB_color = mRGBA_color + (1 - alpha) x background`
    """
    b, g, r, a = cv2.split(bgra)
    if (ia := ~a).any():
        # rgb[:] += rgb_base * ia // 255
        for dst, lut in zip([r, g, b], _LUT_CA[rgb_base]):
            dst[:] += cv2.LUT(ia, lut)
    return cv2.merge([r, g, b])


# def _mbgra_to_rgb(bgra: np.ndarray, rgb_base: np.ndarray) -> np.ndarray:
#     """
#     Pre-multiplied BGRA to RGB. Only image colors are kept.
#     I.e. `RGB_color = alpha ? (mRGBA_color / alpha) : background`
#     """
#     bgra = cv2.cvtColor(bgra, cv2.COLOR_mRGBA2RGBA)  # alpha unscale
#     rgb, a = bgra[..., 2::-1], bgra[..., 3]
#     rgb[a == 0] = rgb_base  # fill fully transparent pixels
#     return rgb


@dataclass(frozen=True, kw_only=True)
class _Image(Image):
    name: bytes
    osd: 'Openslide'

    def numpy(self) -> np.ndarray:
        bgra = np.empty((*self.shape[:2], 4), 'u1')
        OSD.openslide_read_associated_image(
            self.osd.ptr, self.name, c_void_p(bgra.ctypes.data)
        )
        rgb = _mbgra_to_rgb(bgra, self.osd.bg_color)
        rgb = np.ascontiguousarray(rgb)
        return self._postprocess(rgb)

    @property
    def icc(self) -> Icc | None:
        if not _OSD_4X:
            raise NotImplementedError(
                f'Openslide 4.x is required. Got {_VERSION}'
            )

        size = OSD.openslide_get_associated_image_icc_profile_size(
            self.osd.ptr, self.name
        )
        if not size:
            return None

        buf = np.empty(size, 'u1')
        OSD.openslide_read_associated_image_icc_profile(
            self.osd.ptr, self.name, c_void_p(buf.ctypes.data)
        )
        return Icc(buf.tobytes())


@dataclass(frozen=True, kw_only=True)
class _Level(PartsMixin, ImageLevel):
    downsample: int
    index: int
    osd: 'Openslide'

    def part(self, *loc: Span) -> np.ndarray:
        box, valid_box, shape = self._unpack_2d_loc(*loc)

        (y0, y1), (x0, x1) = valid_box
        if y0 == y1 or x0 == x1:  # Patch is outside slide
            return np.broadcast_to(self.osd.bg_color, (*shape, 3))

        bgra = np.empty((y1 - y0, x1 - x0, 4), dtype='u1')
        OSD.openslide_read_region(
            self.osd.ptr,
            c_void_p(bgra.ctypes.data),
            int(x0 * self.downsample),
            int(y0 * self.downsample),
            self.index,
            int(x1 - x0),
            int(y1 - y0),
        )
        rgb = _mbgra_to_rgb(bgra, self.osd.bg_color)

        rgb = self._expand(rgb, valid_box, box, self.osd.bg_color)
        return self._postprocess(rgb)

    @property
    def icc(self) -> Icc | None:
        return self.osd.icc


class Openslide(Driver):
    def __init__(self, path: str) -> None:
        self.ptr = OSD.openslide_open(path.encode())
        if not self.ptr:
            raise ValueError('libopenslide failed to open file')

        if err := OSD.openslide_get_error(self.ptr):
            raise ValueError(err)
        weakref.finalize(self, OSD.openslide_close, self.ptr)

        if _CACHE is not None:  # Use global tile cache for all slides
            OSD.openslide_set_cache(self.ptr, _CACHE)

        meta = self._get_metadata()
        self.meta = unflatten(meta)
        self.osd_meta: dict = self.meta.get('openslide', {})

        bg_hex = self.osd_meta.get('background-color', 'FFFFFF')
        self.bg_color: np.ndarray = np.frombuffer(bytes.fromhex(bg_hex), 'u1')

    def get_mpp(self) -> float | None:
        mpp = (self.osd_meta.get(f'mpp-{t}') for t in ('y', 'x'))
        if s := [float(m) for m in mpp if m]:
            return float(np.mean(s))

        tiff_meta = self.meta.get('tiff', {})
        unit: str = tiff_meta.get('ResolutionUnit', '')
        unit_base = {'centimeter': 10_000, 'inch': 2_540}.get(unit)
        if unit_base is None:
            return None

        resolution = (tiff_meta.get(f'{t}Resolution') for t in ('Y', 'X'))
        if s := [(unit_base / float(v)) for v in resolution if v]:
            return float(np.mean(s))

        return None

    def __repr__(self) -> str:
        return f'{type(self).__name__}({addressof(self.ptr.contents):0x})'

    def _get_metadata(self) -> dict[str, str]:
        names = OSD.openslide_get_property_names(self.ptr)
        r = {}
        for name in _ntas_to_iter(names):
            if value := OSD.openslide_get_property_value(self.ptr, name):
                try:
                    r[name.decode()] = value.decode()
                except UnicodeDecodeError:
                    continue
        return r

    def __len__(self) -> int:
        return OSD.openslide_get_level_count(self.ptr)

    def __getitem__(self, index: int) -> _Level:
        h, w = c_int64(), c_int64()
        OSD.openslide_get_level_dimensions(self.ptr, index, byref(w), byref(h))
        downsample = OSD.openslide_get_level_downsample(self.ptr, index)
        if downsample <= 0:
            raise ValueError(f'Invalid level downsample: {downsample}')

        return _Level(
            (h.value, w.value, 3),
            downsample=round2(downsample),
            index=index,
            osd=self,
        )

    def keys(self) -> list[str]:
        names = OSD.openslide_get_associated_image_names(self.ptr)
        return [name.decode() for name in _ntas_to_iter(names)]

    def get(self, key: str) -> _Image:
        name = key.encode()
        w, h = c_int64(), c_int64()
        OSD.openslide_get_associated_image_dimensions(
            self.ptr, name, byref(w), byref(h)
        )
        return _Image((h.value, w.value, 3), key=key, name=name, osd=self)

    @property
    def bbox(self) -> tuple[slice, ...]:
        bbox = (
            self.osd_meta.get(f'bounds-{t}')
            for t in ('y', 'x', 'height', 'width')
        )
        y0, x0, h, w = (
            int(s) if s and s.strip().isdigit() else None for s in bbox
        )
        y1, x1 = (
            o + s if o is not None and s is not None else None
            for o, s in [(y0, h), (x0, w)]
        )
        return slice(y0, y1), slice(x0, x1)

    @cached_property
    def icc(self) -> Icc | None:
        if not _OSD_4X:
            raise NotImplementedError(
                f'Openslide 4.x is required. Got {_VERSION}'
            )

        size = OSD.openslide_get_icc_profile_size(self.ptr)
        if not size:
            return None

        buf = np.empty(size, 'u1')
        OSD.openslide_read_icc_profile(self.ptr, c_void_p(buf.ctypes.data))
        return Icc(buf.tobytes())



"""
Pure python driver inspired by libtiff
- fast
- thread safe
- compatible with different variants of TIFF like AperioSVS, Ventana, e.t.c.
- not compatible with Hamamatsu NDPI
"""

__all__ = ['Tiff']

import struct
import sys
import warnings
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field, replace
from datetime import datetime
from enum import Enum
from functools import partial
from itertools import chain, product
from math import gcd
from threading import Lock
from typing import Any, Self, assert_never

import cv2
import imagecodecs
import numpy as np
import numpy.typing as npt
from glow import shared_call, si_bin, starmap_n

from bipl._env import env
from bipl._fileio import Paged, fopen
from bipl._types import NDIndex, Patch, Shape, Span

from ._slide_bases import Driver, Image, ImageLevel, PartMixin, ProxyLevel
from ._util import (
    Icc,
    get_aperio_properties,
    get_ventana_properties,
    parse_xml,
    unflatten,
)

_U8 = npt.NDArray[np.uint8]
_U64 = npt.NDArray[np.uint64]

_RESOLUTION_UNITS = {2: 25400, 3: 10000}


# ---------------------------- TIFF tags mappings ----------------------------
class _ColorSpace(Enum):
    MINISBLACK = 1
    RGB = 2
    YCBCR = 6


def nv12_to_rgb(x: _U8) -> _U8:
    w = x.shape[1]
    assert w % 2 == 0
    w2 = w // 2

    # h/2 w 3 -> h/2 w/2 6
    # 6 channels are [Y-Y-Y-Y-Cb-Cr] of 2x2 pixel block
    hw6 = x.reshape(-1, w2, 6)

    # -> h/2 w/2 4 -> h/2 w/2 2 2 -> h/2 2 w/2 2 -> h w
    y = hw6[..., :4].reshape(-1, w2, 2, 2).transpose(0, 2, 1, 3).reshape(-1, w)
    # -> h/2 w/2 2
    cb_cr = hw6[:, :, 4:]

    r = cv2.cvtColorTwoPlane(y, cb_cr, cv2.COLOR_YUV2RGB_NV12)
    return np.asarray(r, 'u1')


class _Planarity(Enum):
    CONTIG = 1  # Single buffer for all image planes
    # SEPARATE = 2  # Separate buffers for each R/G/B plane. TODO: support


class _Orientation(Enum):
    # fmt: off
    TOP_LEFT = 1      # [y, x]
    TOP_RIGHT = 2     # [y, -x] -> im[:, ::-1]
    BOTTOM_RIGHT = 3  # [-y, -x] -> im[::-1, ::-1]
    BOTTOM_LEFT = 4   # [-y, x]  -> im[::-1]
    LEFT_TOP = 5      # [x, y] -> im.transpose(1, 0, 2)
    RIGHT_TOP = 6     # [-x, y] -> im.transpose(1, 0, 2)[:, ::-1]
    RIGHT_BOTTOM = 7  # [-x, -y] -> im.transpose(1, 0, 2)[::-1, ::-1]
    LEFT_BOTTOM = 8   # [x, -y] -> im.transpose(1, 0, 2)[::-1]
    # fmt: on


class _Tag(Enum):
    # &1 = reduced resolution, &2 = page of multipage, &4 = transparency mask
    NEW_SUBFILE_TYPE = 254
    IMAGE_WIDTH = 256
    IMAGE_HEIGHT = 257
    BITS_PER_SAMPLE = 258
    COMPRESSION = 259
    COLORSPACE = 262
    DESCRIPTION = 270
    MAKE = 271
    MODEL = 272
    STRIP_OFFSETS = 273
    ORIENTATION = 274
    SAMPLES_PER_PIXEL = 277
    STRIP_HEIGHT = 278
    STRIP_NBYTES = 279
    RES_X = 282
    RES_Y = 283
    PLANAR = 284
    RES_UNIT = 296
    SOFTWARE = 305
    DATETIME = 306
    ARTIST = 315
    PREDICTOR = 317
    TILE_WIDTH = 322
    TILE_HEIGHT = 323
    TILE_OFFSETS = 324
    TILE_NBYTES = 325
    # SUBIFDS = 330  # offset to child IFDs
    SAMPLE_FORMAT = 339
    SAMPLE_MIN = 340
    SAMPLE_MAX = 341
    JPEG_TABLES = 347
    BACKGROUND_COLOR = 434
    YUV_COEFFICIENTS = 529
    YUV_SUBSAMPLING = 530
    YUV_POSITIONING = 531
    REF_BLACK_WHITE = 532
    XMP = 700
    IMAGE_DEPTH = 32997
    CZ_LSMINFO = 34412
    ICC_PROFILE = 34675
    # EER metadata
    BITS_SKIP_POS = 65007
    BITS_HORZ_SUB = 65008
    BITS_VERT_SUB = 65009
    # NDPI stiff
    NDPI_VERSION = 65420  # 1 for any NDPI
    SOURCE_LENS = 65421  # macro = -1, map of non-empty regions = -2
    OFFSET_X = 65422
    OFFSET_Y = 65423
    OFFSET_Z = 65424
    # = 65425
    # = 65426  # Low 32 bits of optimisation entries
    REFERENCE = 65427
    AUTH_CODE = 65428
    # = 65432  # High 32 bits of optimisation entries
    # = 65433
    # = 65439
    # = 65440
    # = 65441  # 0?
    SCANNER_SERIAL_NUMBER = 65442
    # = 65443  # 0 or 16?
    # = 65444  # 80?
    # = 65445  # 0, 2 or 10?
    # = 65446  # 0?
    # = 65449  # ASCII metadata, key=value pairs
    # = 65455  # 13?
    # = 65456  # 101?
    # = 65457  # 0?
    # = 65458  # 0?


class _ImageFileDirectory(dict[_Tag, Any]):
    def __init__(self, tags: dict[_Tag, Any], index: int) -> None:
        # Use strong types
        for k, t in [
            (_Tag.COLORSPACE, _ColorSpace),
            (_Tag.ORIENTATION, _Orientation),
            (_Tag.PLANAR, _Planarity),
            (
                _Tag.DATETIME,
                lambda x: datetime.strptime(x, '%Y:%m:%d %H:%M:%S'),
            ),
            (_Tag.PREDICTOR, _unpredictors.get),
            (_Tag.REF_BLACK_WHITE, lambda x: x.reshape(3, 2)),
            (_Tag.ICC_PROFILE, Icc),
        ]:
            if (v := tags.get(k)) is not None:
                tags[k] = t(v)

        super().__init__(tags)
        self.index = index

        if (
            self.get(_Tag.NDPI_VERSION) == 1
            or self.get(_Tag.MAKE) == 'Hamamatsu'
        ):
            raise ValueError('TIFF: Hamamatsu is not yet supported')
        # ! unused
        self.bps = self.get(_Tag.BITS_PER_SAMPLE, 1)
        self.sample_format = self.get(_Tag.SAMPLE_FORMAT, 1)

        self.subsampling = (1, 1)

        # TODO: use this in YCbCr conversion
        self.gray = np.array([], 'f4')  # Luma coefficients
        self.yuv_centered = True
        self.yuv_bw = np.array([], 'f4')  # BW pairs, per channel

        if self[_Tag.COLORSPACE] is _ColorSpace.YCBCR:
            self.gray = np.array([0.299, 0.587, 0.114], 'f4')
            self.gray = self.get(_Tag.YUV_COEFFICIENTS, self.gray)

            # YCbCr subsampling H/W, i.e:
            # 24bps:
            #  (1 1) = 4:4:4 (cccc/cccc), 100% H, 100% W
            # 16bps:
            #  (1 2) = 4:2:2 (c_c_/c_c_), 100% H, 50% W
            #  (2 1) = 4:4:0 (cccc/____), 50% H, 100% W
            # 12bps:
            #  (1 4) = 4:1:1 (c___/c___), 100% H, 25% W
            #  (2 2) = 4:2:0 (c_c_/____), 50% H, 50% W - a.k.a. YUV-NV12
            # 10bps:
            #  (2 4) = 4:1:0 (c___/____), 50% H, 25% W
            ss_w, ss_h = self.get(_Tag.YUV_SUBSAMPLING, (2, 2))
            self.subsampling = ss_h, ss_w

            self.yuv_centered = self.get(_Tag.YUV_POSITIONING) != 2
            self.yuv_bw = self.get(_Tag.REF_BLACK_WHITE, self.yuv_bw)

        self.image_shape = (
            self[_Tag.IMAGE_HEIGHT],
            self[_Tag.IMAGE_WIDTH],
            self[_Tag.SAMPLES_PER_PIXEL],
        )
        self.is_tiled, self.unit_shape, self.spans = self._tiling_schema()

    def _tiling_schema(self) -> tuple[bool, Shape, _U64]:
        """
        Returns tile shape &
        (h w d 2) tensor of start:stop of each tile w.r.t file.
        """
        spp = self[_Tag.SAMPLES_PER_PIXEL]

        if th := self.get(_Tag.TILE_HEIGHT, 0):  # Tiles
            tw = self[_Tag.TILE_WIDTH]
            offset_tag = _Tag.TILE_OFFSETS
            byte_tag = _Tag.TILE_NBYTES
            is_tiled = True

        else:  # Strips
            th = self[_Tag.STRIP_HEIGHT]
            tw = self[_Tag.IMAGE_WIDTH]
            offset_tag = _Tag.STRIP_OFFSETS
            byte_tag = _Tag.STRIP_NBYTES
            is_tiled = False

        unit_shape = (th, tw, spp)
        grid_shape = tuple(
            len(range(0, s, t)) for s, t in zip(self.image_shape, unit_shape)
        )
        tbs: _U64 = np.reshape(self[offset_tag], grid_shape)
        tbc: _U64 = np.reshape(self[byte_tag], grid_shape)

        grid = np.stack((tbs, tbs + tbc), -1)

        return is_tiled, unit_shape, grid

    def get_decoder(self) -> Callable[[bytes], _U8]:
        compression = self[_Tag.COMPRESSION]
        try:
            decompress = _decompressors[compression]
        except KeyError:
            raise ValueError(
                f'IFD: {self.index}: Unknown compression {compression}'
            ) from None

        if decompress is None:  # RAW
            tile_size = np.prod(self.unit_shape)
            self.spans[..., 1] = self.spans[..., 0] + tile_size

        if decompress in {imagecodecs.jpeg_decode, imagecodecs.jpeg8_decode}:
            return partial(
                decompress,
                tables=self.get(_Tag.JPEG_TABLES),
                colorspace=self[_Tag.COLORSPACE].name,
                outcolorspace='RGB',
            )

        if decompress == imagecodecs.eer_decode:
            match compression:
                case 65001:
                    rlebits, horzbits, vertbits = 8, 2, 2
                case 65002:
                    rlebits, horzbits, vertbits = 7, 2, 2
                case 65002:
                    rlebits = self.get(_Tag.BITS_SKIP_POS, 7)
                    horzbits = self.get(_Tag.BITS_HORZ_SUB, 2)
                    vertbits = self.get(_Tag.BITS_VERT_SUB, 2)
                case _ as unreachable:
                    assert_never(unreachable)

            return partial(
                imagecodecs.eer_decode,
                shape=(self.unit_shape[0], self.unit_shape[1]),
                rlebits=rlebits,
                horzbits=horzbits,
                vertbits=vertbits,
                superres=False,
            )

        if decompress is _jetraw_decode:
            return partial(_jetraw_decode, shape=self.unit_shape)

        if decompress in {
            imagecodecs.jpeg2k_decode,
            imagecodecs.jpegxl_decode,
            imagecodecs.jpegxr_decode,
            imagecodecs.png_decode,
            imagecodecs.webp_decode,
        }:
            return decompress

        unpredict: Callable | None = self.get(_Tag.PREDICTOR)
        return _LambdaDecoder(decompress, unpredict, self.unit_shape)

    def _bg_color(self, meta: dict) -> _U8:
        if c := meta.get('ScanWhitePoint'):
            return np.array(int(c), 'u1')

        bg_hex: bytes = self.get(_Tag.BACKGROUND_COLOR, b'FFFFFF')
        return np.frombuffer(bytes.fromhex(bg_hex.decode()), 'u1').copy()

    def vendor_props(self, vendor: str) -> tuple[str, dict[str, str]] | None:
        description = self.get(_Tag.DESCRIPTION, '')
        xmp = self.get(_Tag.XMP)
        assert isinstance(description, str)

        match vendor:
            case 'ventana':  # BIF of Roche
                if xmp is not None and (
                    meta := get_ventana_properties(xmp, self.index)
                ):
                    return description, meta
                if self.index != 0:
                    return description, {}

            case 'aperio':  # SVS
                return get_aperio_properties(description, self.index)

            case 'generic':  # TIFF
                try:
                    meta = unflatten(parse_xml(description))
                except Exception:  # noqa: BLE001
                    return description, {}
                else:
                    return '', meta

        return None

    def _get_mpp(self, vendor: str, meta: dict) -> float | None:
        """Extract MPP, um/pixel, phisical pixel size"""
        if vendor == 'ventana':
            # Ventana messes with TIFF tag XResolution, YResolution
            return float(mpp_s) if (mpp_s := meta.get('ScanRes')) else None

        res_unit_kind = self.get(_Tag.RES_UNIT)
        if res_unit_kind and (
            res_unit := _RESOLUTION_UNITS.get(res_unit_kind)
        ):
            mpp_xy = [
                res_unit / ppu
                for t in (_Tag.RES_Y, _Tag.RES_X)
                if (ppu := self.get(t))
            ]
            if mpp_xy and (mpp := float(np.mean(mpp_xy))):
                return mpp

        if mpp_s := meta.get('MPP'):
            return float(mpp_s)
        return None

    def properties(self) -> tuple[str, str, dict[str, str], float | None, _U8]:
        """Extracts: (vendor, header, metadata, mpp, bg color)"""
        for vendor in ('ventana', 'aperio', 'generic'):
            if r := self.vendor_props(vendor):
                header, meta = r
                break
        else:
            raise ValueError('Unknown vendor')

        mpp = self._get_mpp(vendor, meta)
        bg_color = self._bg_color(meta)
        return vendor, header, meta, mpp, bg_color


# ------------------ image, level & opener implementations -------------------


@dataclass(frozen=True, slots=True)
class _LambdaDecoder:
    decompress: Callable[[bytes], bytes] | None
    unpredict: Callable | None
    shape: Shape

    def __call__(self, buf: bytes) -> _U8:
        if self.decompress:
            buf = self.decompress(buf)

        h, w, c = self.shape
        arr = np.frombuffer(buf, 'u1')
        arr = arr[: h * w * c].reshape(-1, w, c)

        if self.unpredict:
            arr = self.unpredict(arr, axis=-2, out=arr)

        if arr.shape[0] == h:
            return arr
        return np.pad(arr, ((0, h - arr.shape[0]), (0, 0), (0, 0)))


def _jetraw_decode(buf: bytes, *, shape: Shape) -> _U8:
    out = np.zeros(shape, 'u2')
    imagecodecs.jetraw_decode(buf, out=out.ravel())
    return out


# For IDs see bipl.io._tiff_compressions
_decomp_to_ids: dict[Callable | None, set[int]] = {
    # bytes -> bytes
    None: {1},
    imagecodecs.deflate_decode: {8, 32946, 50013},
    imagecodecs.lzma_decode: {34925},
    imagecodecs.lzw_decode: {5},
    imagecodecs.packbits_decode: {32773},
    imagecodecs.zstd_decode: {34926, 50000},
    # bytes -> ndarray, no predictor
    _jetraw_decode: {48124},
    imagecodecs.eer_decode: {65000, 65001, 65002},
    imagecodecs.jpeg_decode: {6, 7, 33007},
    imagecodecs.jpeg2k_decode: {33003, 33004, 33005, 34712},
    imagecodecs.jpeg8_decode: {34892},
    imagecodecs.jpegxl_decode: {50002, 52546},
    imagecodecs.jpegxr_decode: {22610, 34934},
    imagecodecs.lerc_decode: {34887},
    imagecodecs.png_decode: {34933},
    imagecodecs.webp_decode: {34927, 50001},
}
_decompressors: dict[int, Callable | None] = {}
for _decomp, _codec_ids in _decomp_to_ids.items():
    for _codec_id in _codec_ids:
        assert _codec_id not in _decompressors, f'Duplicates: {_decomp_to_ids}'
        _decompressors[_codec_id] = _decomp


_unpredictors: dict[int, Callable] = {
    2: imagecodecs.delta_decode,
    3: imagecodecs.floatpred_decode,
    34892: partial(imagecodecs.delta_decode, dist=2),
    34893: partial(imagecodecs.delta_decode, dist=4),
    34894: partial(imagecodecs.floatpred_decode, dist=2),
    34895: partial(imagecodecs.floatpred_decode, dist=4),
}


@dataclass(eq=False, frozen=True, kw_only=True)
class _BaseImage(Image):
    icc_impl: Icc | None = None
    buf: Paged
    path: str
    spans: _U64 = field(repr=False)
    tile_shape: Shape
    decode: Callable[[bytes], _U8]
    cache: '_CacheZYXC'
    fill: _U8
    decimations: int = 0  # [0, 1, 2, ..., n]
    prev: ImageLevel | None = None

    @property
    def icc(self) -> Icc | None:
        return self.icc_impl

    def decimate(self, dst: float, src: int = 1) -> tuple[int, Self]:
        assert dst >= 1
        assert src >= 1
        th, tw, tc = self.tile_shape
        t = gcd(th, tw)

        max_ds = (-t) & t  # See: https://stackoverflow.com/q/1551775
        steps = min(int(dst / src), max_ds).bit_length() - 1

        if steps <= 0:  # No need or cannot decimate
            return (src, self)

        ds = 2**steps
        h, w, c = self.shape

        prev = self.prev
        shape = ((h + ds - 1) // ds, (w + ds - 1) // ds, c)

        if prev is not None:
            prev_ds, prev = prev.decimate(ds, 1)
            if prev_ds != ds:
                prev = prev.rescale(prev_ds / ds)

        lv = replace(
            self,
            shape=shape,
            tile_shape=(th // ds, tw // ds, tc),
            decimations=self.decimations + steps,
            prev=prev,
        )
        return (src * ds, lv)

    def __eq__(self, rhs) -> bool:  # Used by _Level.tile:shared_call
        return (
            type(self) is type(rhs)
            and self.buf is rhs.buf
            and self.shape == rhs.shape
            and self.decimations == rhs.decimations
        )

    def __hash__(self) -> int:  # Used by _Level.tile:shared_call
        return hash((self.buf, self.shape, self.decimations))

    @shared_call  # Thread safety
    def tile(self, *idx: int) -> _U8 | None:
        lo, hi = self.spans[idx].tolist()
        if lo == hi:  # If nothing to read, don't read
            return None

        # NOTE:
        # Like OpenSlide does we use private cache for each slide
        #   with (level, y, x) key to cache decoded pixels.
        # But we also cache opened slides to not waste time on re-opening
        #   (that can lead to multiple caches existing at the same moment).
        key = (self.decimations, lo, hi)

        # Cache hit
        if (im := self.cache[key]) is not None:
            return self._postprocess(im)

        # Cache miss
        # Read tile from disk
        im = self.decode(self.buf.pread(hi - lo, lo))

        # Resize if level is pooled
        th, tw = self.tile_shape[:2]
        if self.decimations == 1:
            im = cv2.resize(im, (tw, th))
        elif self.decimations >= 2:
            im = cv2.resize(im, (tw, th), interpolation=cv2.INTER_AREA)

        self.cache[key] = im
        return self._postprocess(im)

    def parts(
        self, locs: Sequence[tuple[Span, ...]], max_workers: int = 0
    ) -> Iterator[Patch]:
        if not locs:
            return
        n = len(locs)

        # (n yx lo/hi)
        boxes = np.asarray(locs, 'i4')
        th, tw, spp = self.tile_shape

        # (n yxc)
        out_shapes = np.empty((n, 3), int)
        out_shapes[:, :2] = boxes @ [-1, 1]
        out_shapes[:, 2] = spp

        # tile idx -> [(loc id, o loc, t loc), ...]
        tile_to_boxes: dict[
            NDIndex,
            list[tuple[int, tuple[slice, ...], tuple[slice, ...]]],
        ] = {}

        counts = np.zeros(n, int)  # num used tiles
        box_ids, mins_, vboxes, i_lows, i_highs, counts_, o_locs = (
            self._init_index(boxes)
        )
        counts[box_ids] = counts_
        for i, *args in zip(box_ids.tolist(), mins_, vboxes, i_lows, i_highs):
            for iyx, oyx, tyx in self._make_subindex(*args):
                tile_to_boxes.setdefault((*iyx, 0), []).append((i, oyx, tyx))

        # box idx -> patch, buffer to store results before they're complete
        fill = self._postprocess(self.fill[None, None])[0, 0, :]
        buf: dict[int, Patch] = {
            i: Patch(loc, np.broadcast_to(fill, oshape))
            for i, (loc, oshape) in enumerate(zip(locs, out_shapes))
        }
        # Pop initial nulls
        pos = box_ids[0] if box_ids.size else n
        for i in range(pos):
            yield buf.pop(i)
        if not buf:
            return

        # Read linear according to file layout
        ids: list[NDIndex] = []
        nulls: list[NDIndex] = []
        for iyx in tile_to_boxes:
            nbytes = self.spans[iyx] @ [-1, 1]
            (ids, nulls)[int(nbytes) == 0].append(iyx)

        prev_tiles: Iterator[tuple[NDIndex, np.ndarray | None]]
        if self.prev is not None and nulls:
            prev_parts = self.prev.parts(
                [
                    ((iy * th, iy * th + th), (ix * tw, ix * tw + tw))
                    for iy, ix, _ in nulls
                ],
                max_workers=max_workers,
            )
            prev_tiles = ((i, a) for i, (_, a) in zip(nulls, prev_parts))
        else:
            prev_tiles = ((i, None) for i in nulls)

        ids = sorted(ids, key=self.spans[:, :, :, 0].__getitem__)
        tiles = zip(ids, starmap_n(self.tile, ids, max_workers=max_workers))

        # Unwrap & build patches
        rois: dict[int, tuple[Span, ...]] = dict(
            zip(box_ids.tolist(), o_locs.tolist())
        )
        for iyx, tile in chain(prev_tiles, tiles):

            # Update all parts using current tile
            for i, oyx, tyx in tile_to_boxes.get(iyx, []):
                out = buf[i]

                if roi := rois.pop(i, None):
                    a = np.empty_like(out.data)
                    (y0, y1), (x0, x1) = roi
                    a[:y0] = fill
                    a[y0:y1, :x0] = fill
                    a[y0:y1, x1:] = fill
                    a[y1:] = fill
                    buf[i] = out = Patch(out.loc, a)

                out.data[oyx] = fill if tile is None else tile[tyx]
                counts[i] -= 1

            # Yield completed items, order preserved
            while pos < n and counts[pos] == 0:
                yield buf.pop(pos)
                pos += 1

    def _init_index(self, boxes: np.ndarray) -> tuple[  # (n yx lo/hi)
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]:
        shape = np.reshape(self.shape[:2], (2, 1))
        tile_size = np.reshape(self.tile_shape[:2], (1, 2))

        # (n yx lo/hi), box within slide
        vboxes = boxes.clip(0, shape)

        # (n yx)
        mins_ = boxes[:, :, 0]
        box_sizes = boxes @ [-1, 1]
        vbox_sizes = vboxes @ [-1, 1]

        # (n), drop empty "valid" boxes
        mask = vbox_sizes.all(-1)
        [box_ids] = mask.nonzero()
        mins_, box_sizes, vboxes = mins_[mask], box_sizes[mask], vboxes[mask]

        # (n yx)
        i_lows = vboxes[:, :, 0] // tile_size  # i_lo * tile <= vmin
        i_highs = -(-vboxes[:, :, 1] // tile_size)  # i_hi * tile >= vmax

        # (n)
        counts = (i_highs - i_lows).prod(-1)
        assert counts.all()

        # (n yx lo/hi)
        o_locs = vboxes - mins_[:, :, None]  # region of `out` to fill
        o_locs = o_locs.clip(0, box_sizes[:, :, None])

        # (n) (n yx) (n yx lo/hi) (n yx) (n yx) (n) {i -> (yx lo/hi)}
        return box_ids, mins_, vboxes, i_lows, i_highs, counts, o_locs

    def _make_subindex(
        self,
        min_: np.ndarray,  # (yx)
        vbox: np.ndarray,  # (yx lo/hi)
        i_lo: np.ndarray,  # (yx)
        i_hi: np.ndarray,  # (yx)
    ) -> list[tuple[NDIndex, tuple[slice, ...], tuple[slice, ...]]]:
        # (yx)
        tile_size = self.tile_shape[:2]

        # (ny) (nx)
        ids = map(range, i_lo, i_hi)
        oxs: list[list[slice]] = []
        txs: list[list[slice]] = []
        for i_lo_, i_hi_1, tsize, vspan, amin_ in zip(
            i_lo, i_hi + 1, tile_size, vbox, min_
        ):
            # (n + 1)
            ts = np.arange(i_lo_, i_hi_1) * tsize
            vts = ts.clip(*vspan)
            vts_src = vts - amin_

            # (n), result slices
            o_crops = map(slice, vts_src[:-1], vts_src[1:])
            oxs.append(list(o_crops))

            # (n), tile slices
            t_crops = map(slice, vts[:-1] - ts[:-1], vts[1:] - ts[:-1])
            txs.append(list(t_crops))

        return [*zip(product(*ids), product(*oxs), product(*txs), strict=True)]

    def apply(self, fn: Callable[[np.ndarray], np.ndarray]) -> Self:
        post = [*self.post, fn]

        prev = self.prev
        if prev is not None:
            prev = replace(prev, post=post)

        return replace(self, post=post, prev=prev)


def _get_key(
    ifd: _ImageFileDirectory, vendor: str, head: str, index: int
) -> str | None:
    if ifd.is_tiled:
        return None

    match (vendor, index, head):
        case ('aperio', 1, _) | ('ventana', _, 'Thumbnail'):
            return 'thumbnail'
        case ('ventana', _, 'Label Image') | ('ventana', _, 'Label_Image'):
            return 'macro'
        case ('ventana', _, _):
            return None

    for key in ('label', 'macro'):
        if any(key in s for s in head.splitlines()):
            return key
    return None


@dataclass(eq=False, frozen=True, kw_only=True)
class _Image(_BaseImage):
    def __eq__(self, rhs) -> bool:  # Used by _Level.tile:shared_call
        return (
            type(self) is type(rhs)
            and self.buf is rhs.buf
            and self.shape == rhs.shape
            and self.decimations == rhs.decimations
            and self.key == rhs.key
        )

    def __hash__(self) -> int:  # Used by _Level.tile:shared_call
        return hash((self.buf, self.shape, self.decimations, self.key))

    def numpy(self) -> np.ndarray:
        """Retrieve whole image as array"""
        h, w, _ = self.shape
        [(_, a)] = self.parts([((0, int(h)), (0, int(w)))])
        return a


@dataclass(eq=False, frozen=True, kw_only=True)
class _Level(PartMixin, ImageLevel, _BaseImage):
    def __eq__(self, rhs) -> bool:  # Used by _Level.tile:shared_call
        return (
            type(self) is type(rhs)
            and self.buf is rhs.buf
            and self.shape == rhs.shape
            and self.decimations == rhs.decimations
        )

    def __hash__(self) -> int:  # Used by _Level.tile:shared_call
        return hash((self.buf, self.shape, self.decimations))


class Tiff(Driver):
    def __init__(self, path: str) -> None:
        self._buf = fopen(path)
        self._path = path
        self.cache = _CacheZYXC()
        self._ifds = list(self._iter_ifds())

        # Directory 0 is main
        self.vendor, self._head, self._meta, self.mpp, self._bg_color = (
            self._ifds[0].properties()
        )

    def _iter_ifds(self) -> Iterator[_ImageFileDirectory]:
        header = self._buf.pread(4, 0)
        try:
            byteorder = {b'II': '<', b'MM': '>', b'EP': '<'}[header[:2]]
        except KeyError as exc:
            raise ValueError(f'not a TIFF file {header!r}') from exc
        assert byteorder in ('<', '>')

        [version] = struct.unpack(byteorder + 'H', header[2:4])
        if version == 42:  # TIFF
            usize = 4
            head_fmt = 'H'  # u2
            head_size = 2
            o_fmt = 'I'  # u4
        elif version == 43:  # BigTIFF
            usize, zero = struct.unpack(
                byteorder + 'HH', self._buf.pread(4, 4)
            )
            if usize != 8 or zero != 0:
                raise ValueError(
                    f'invalid BigTIFF offset size {(usize, zero)}'
                )
            head_fmt = o_fmt = 'Q'  # u8
            head_size = 8
        else:
            raise ValueError(f'invalid TIFF version: {version}')

        t_body = np.dtype(
            [
                ('tag', 'H'),
                ('type', 'H'),
                ('count', o_fmt),
                ('value', f'{usize}B'),
            ]
        )
        dtypes = _DTYPES
        if sys.byteorder != {'<': 'little', '>': 'big'}[byteorder]:
            head_fmt = byteorder + head_fmt
            o_fmt = byteorder + o_fmt
            t_body = t_body.newbyteorder()
            dtypes = {i: dt.newbyteorder() for i, dt in dtypes.items()}

        [pos] = struct.unpack(o_fmt, self._buf.pread(usize, usize))
        idx = 0
        while pos:
            # Read IFD header
            [num_tags] = struct.unpack(
                head_fmt, self._buf.pread(head_size, pos)
            )
            pos += head_size

            # Read tags
            ts = np.frombuffer(
                self._buf.pread(t_body.itemsize * num_tags, pos), t_body
            )
            pos += ts.nbytes

            tags = self._get_ifd_tags(ts, o_fmt=o_fmt, dtypes=dtypes)
            yield _ImageFileDirectory(tags, idx)

            # Jump to next IFD
            [pos] = struct.unpack(o_fmt, self._buf.pread(usize, pos))
            idx += 1

    def _get_ifd_tags(
        self, ts: np.ndarray, o_fmt: str, dtypes: Mapping[int, np.dtype]
    ) -> dict[_Tag, Any]:
        m = np.isin(ts['tag'], _TAG_NAMES_A) & np.isin(ts['type'], _DTYPES_A)
        if (unknown := ts[~m][['tag', 'type']]).size:
            unknown = {t: _DTYPES.get(i, i) for t, i in unknown}
            raise ValueError(f'Unknown tags: {unknown}')
        ts = ts[m]

        tags: dict[_Tag, Any] = {}
        for tag_, type_, count, value in ts:
            dt = dtypes[type_]
            size = int(count * dt.itemsize)

            if size > struct.calcsize(o_fmt):
                [o] = struct.unpack(o_fmt, value)
                v = np.frombuffer(self._buf.pread(size, o), dt.base)
            else:
                v = value[:size].view(dt.base)

            if dt.shape:  # 2u4, 2i4
                v = v[::2] / v[1::2]

            if type_ == 2:  # char
                v = v.tobytes().removesuffix(b'\0')
                try:
                    v = v.decode()
                except UnicodeDecodeError:
                    v = ascii(v)
            elif type_ == 7:  # undefined
                v = v.tobytes()
            elif v.size == 1:
                v = v.item()

            tags[_Tag(tag_)] = v

        return tags

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._path})'

    def get_mpp(self) -> float | None:
        return self.mpp

    def __len__(self) -> int:
        n = len(self._ifds)
        if n < 3:
            raise ValueError(f'Not enough levels (<3): {n}')
        return n

    def __getitem__(self, index: int) -> Image | None:
        ifd = self._ifds[index]
        if index == 0:
            head = self._head
        else:
            r = ifd.vendor_props(self.vendor)
            if not r:
                raise ValueError('File directories are from different vendors')
            head, _ = r

        if ifd[_Tag.COLORSPACE] is _ColorSpace.YCBCR and (
            ifd.subsampling != (2, 2)
        ):
            raise ValueError(f'Unsupported YUV subsampling: {ifd.subsampling}')

        key = _get_key(ifd, self.vendor, head, index)
        tp = _Level if ifd.is_tiled else _Image
        return tp(
            ifd.image_shape,
            key=key,
            icc_impl=ifd.get(_Tag.ICC_PROFILE),
            buf=self._buf,
            path=self._path,
            spans=ifd.spans,
            tile_shape=ifd.unit_shape,
            decode=ifd.get_decoder(),
            cache=self.cache,
            fill=self._bg_color,
        )

    def build_pyramid(
        self, levels: Sequence[ImageLevel]
    ) -> tuple[tuple[int, ...], list[ImageLevel]]:
        downsamples, levels = super().build_pyramid(levels)

        if self.vendor != 'aperio':
            return downsamples, levels

        # Aperio can choke on non-L0 levels. Read those from L0 to fix.
        # `openslide` does this for us, delegate to it.
        for i, (ds, lv) in enumerate(zip(downsamples, levels)):
            assert isinstance(lv, _Level)
            empty = lv.spans[..., 0] == lv.spans[..., 1]

            if i == 0:
                if empty.any():
                    raise ValueError('Found 0s in tile size table')
                continue

            if empty.any():  # Cut tile grid right after first empty tile
                first_empty = empty.argmax()
                lv.spans.reshape(-1, 2)[first_empty:] = 0

            prev = levels[0]
            prev_ds, prev = prev.decimate(ds, 1)

            if prev_ds != ds:
                prev = ProxyLevel(
                    lv.shape,
                    prev.post,
                    scale=prev_ds / ds,
                    base=replace(prev, post=[]),
                )

            levels[i] = replace(lv, prev=prev)

        return downsamples, levels


# ------------------------------- tile caching -------------------------------


class _CacheZYXC:
    __slots__ = ('items', 'lock', 'used')

    def __init__(self) -> None:
        self.lock = Lock()
        self.used: int = 0
        self.items: dict[tuple, tuple[int, _U8]] = {}  # IYXC -> (size, buf)

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}'
            f'(used={si_bin(self.used)}, items={len(self.items)})'
        )

    def __getitem__(self, key: tuple) -> _U8 | None:
        with self.lock:
            if e := self.items.get(key):
                return e[1]
        return None

    def __setitem__(self, key: tuple, obj: _U8) -> None:
        if not (capacity := env.BIPL_TILE_CACHE):
            return
        size = obj.nbytes
        if size > capacity:
            warnings.warn(
                f'Rejecting overlarge cache entry of size {size} bytes',
                stacklevel=3,
            )
            return
        max_size = capacity - size
        with self.lock:
            while self.items and self.used > max_size:
                k = next(iter(self.items))
                self.used -= self.items.pop(k)[0]
            if self.used <= max_size:
                self.items[key] = (size, obj)
                self.used += size


_DTYPES: dict[int, np.dtype] = {
    1: np.dtype('u1'),
    2: np.dtype('u1'),  # Null-terminated ascii string
    3: np.dtype('u2'),
    4: np.dtype('u4'),
    5: np.dtype('2u4'),
    6: np.dtype('i1'),
    7: np.dtype('u1'),  # Undefined
    8: np.dtype('i2'),
    9: np.dtype('i4'),
    10: np.dtype('2i4'),
    11: np.dtype('f4'),
    12: np.dtype('f8'),
    16: np.dtype('u8'),
}

_DTYPES_A = np.array([*_DTYPES], 'u2')
_TAG_NAMES_A = np.array([*_Tag._value2member_map_], 'u2')



"""
Start coverage server via:
    python -m bipl.cov --roodir /slides/anns/roi --host 0.0.0.0 --port 7575

In client set environment variable
    BIPL_COV_URL=http://{server_host}:7575
"""

__all__ = ['update']

import asyncio
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from urllib.request import Request, urlopen

from glow.cli import parse_args
from pydantic import BaseModel

from bipl._env import env
from bipl._types import Shape, Vec


class Update(BaseModel):
    path: str
    y0: int
    y1: int
    x0: int
    x1: int


def update(path: str, z0_yx_offset: Vec, dsize: Shape, scale: float) -> None:
    if not env.BIPL_COV_URL:
        return
    y0, x0 = z0_yx_offset
    y1 = y0 + round(dsize[0] / scale)
    x1 = x0 + round(dsize[1] / scale)

    request = Request(
        f'{env.BIPL_COV_URL}/update',
        method='PUT',
        data=(
            Update(path=path, y0=y0, y1=y1, x0=x0, x1=x1)
            .model_dump_json()
            .encode()
        ),
        headers={'Content-Type': 'application/json'},
    )
    with urlopen(request) as response:
        response.read()


@dataclass
class _Coverage:
    rootdir: Path
    polygons: dict[str, list[list]] = field(default_factory=dict)
    atime: datetime = field(default_factory=datetime.now)
    ndiffs: int = 0
    lock: asyncio.Lock = field(default_factory=asyncio.Lock)

    async def flush(self) -> None:
        import geojson

        for fname, pgs in self.polygons.items():
            path = self.rootdir / f'{fname}.geojson'
            if path.is_file():
                fc: geojson.FeatureCollection = geojson.loads(path.read_text())
                fc[0]['geometry'] += pgs
            else:
                fc = geojson.FeatureCollection(
                    [
                        geojson.Feature(
                            geometry=geojson.MultiPolygon(pgs),
                            properties={'class_name': 'RoI'},
                        )
                    ]
                )
            path.write_text(geojson.dumps(fc))

        self.polygons.clear()

    async def update(self, u: Update) -> None:
        fname = Path(u.path).name
        pg = [[[u.x0, u.y0], [u.x0, u.y1], [u.x1, u.y1], [u.x1, u.y0]]]

        async with self.lock:
            self.polygons.setdefault(fname, []).append(pg)
            self.ndiffs += 1

            # Flush if last flush was long time ago, or lots of changes exist
            now = datetime.now()
            if (self.atime - now).total_seconds() > 60 or self.ndiffs > 1000:
                await self.flush()
                self.atime = now
                self.ndiffs = 0


def _main(rootdir: Path, host: str, port: int = 7575) -> None:
    import uvicorn
    from fastapi import FastAPI

    cov = _Coverage(rootdir=rootdir)
    app = FastAPI()
    app.add_api_route('/update', cov.update, methods=['PUT'], status_code=204)
    uvicorn.run(app, host=host, port=port)


if __name__ == '__main__':
    parse_args(_main)



__all__ = ['LumaScaler', 'Normalizer']

from typing import Literal

import cv2
import numpy as np
from glow import around

LUMA_MAX = 240
LUT_THRESHOLD = 160  # Slide dimmer this considered as low light


class LumaScaler:
    def __init__(self, rgb: np.ndarray) -> None:
        gray = np.asarray(cv2.cvtColor(rgb, cv2.COLOR_RGB2GRAY))
        luma95 = np.percentile(gray, 95)  # type: ignore

        if luma95 <= LUT_THRESHOLD:
            lut = np.arange(256).astype('f4')
            lut *= LUMA_MAX / luma95
            self.lut = around(lut.clip(0, 255), 'u1')
        else:
            self.lut = None

    def __call__(self, rgb: np.ndarray) -> np.ndarray:
        if self.lut is not None:
            return cv2.LUT(rgb.ravel(), self.lut).reshape(rgb.shape)
        return rgb


class Normalizer:
    """
    Normalizes slide's thumbnail & tiles.

    To calibrate use 64um slide level.
    Defaults 64um, r=0.5 and 'ab' channels were acquired during testing.
    """

    def __init__(
        self,
        rgb: np.ndarray | None = None,
        weight: float = 1.0,
        channels: Literal['L', 'ab', 'Lab'] = 'ab',
        r: float = 0.5,
    ) -> None:
        assert 0 < weight <= 1
        self.channels = channels
        self.r = r
        self.weight = weight
        self.lut = None
        if rgb is not None:
            self.lut = self._make_lut(cv2.cvtColor(rgb, cv2.COLOR_RGB2LAB))

    def _make_lut(self, lab: np.ndarray) -> np.ndarray:
        cdf = _minmax_cdf(lab) if self.r == 0 else self._unsharp_cdf(lab)

        if self.weight < 1:
            cdf *= self.weight
            cdf += np.linspace(0, 1 - self.weight, 256, dtype='f4')

        return around(cdf * 255, 'u1')

    def _unsharp_cdf(self, lab: np.ndarray) -> np.ndarray:
        f4 = lab.astype('f4')
        h, w = f4.shape[:2]

        # 256 total value counts
        lstar = lab[:, :, 0].ravel()  # (h w) -> n
        counts = np.bincount(lstar, minlength=256)  # 256

        match self.channels:
            case 'L':
                f4 = f4[:, :, :1]
                f4 *= 1 / 255
            case 'ab':
                f4 = f4[:, :, 1:]
                f4 *= 1 / 100
            case 'Lab':
                f4 *= [1 / 255, 1 / 100, 1 / 100]

        kern = np.array([[1, 2, 1], [2, -12, 2], [1, 2, 1]], 'f4') / 16
        var = cv2.filter2D(f4, -1, kern, borderType=cv2.BORDER_REPLICATE)
        var = var.reshape(h, w, -1)
        var = np.square(var).sum(-1).ravel()  # L2, (h w)
        var = var[lstar.argsort()]

        # 256 variances
        m = counts > 0
        counts_nz = counts[m]
        sep = np.r_[0, counts_nz[:-1]].cumsum()
        df = np.zeros(256, 'f4')
        df[m] = np.add.reduceat(var, sep) / counts_nz  # sum(E^2) -> mean(E^2)
        df **= self.r / 2  # sqrt(mean(E^2)) ^ r
        df /= df.sum()

        return _make_cdf(df)

    def __call__(self, rgb: np.ndarray) -> np.ndarray:
        if not rgb.size:
            return rgb
        lab = cv2.cvtColor(rgb, cv2.COLOR_RGB2LAB)

        if (lut := self.lut) is None:
            lut = self._make_lut(lab)

        lab[:, :, 0] = cv2.LUT(lab[:, :, 0], lut)
        return cv2.cvtColor(lab, cv2.COLOR_LAB2RGB)


def _make_cdf(pdf: np.ndarray) -> np.ndarray:
    cdf = np.ma.masked_equal(pdf.cumsum(), 0).astype('f4')
    cdf -= cdf.min()
    cdf /= cdf.max()
    return cdf.filled(0)


def _minmax_cdf(lab: np.ndarray) -> np.ndarray:
    lstar = lab[:, :, 0]  # (h w)
    lo, hi = int(lstar.min()), int(lstar.max())  # Don't overflow
    return np.r_[
        np.zeros(lo, dtype='f4'),
        np.linspace(0, 1, hi - lo + 1, dtype='f4'),
        np.ones(255 - hi, 'f4'),
    ]



RAW = 1
CCITTRLE = 2  # CCITT 10
CCITT_T4 = 3  # T4/Group 3 Fax
CCITT_T6 = 4  # T6/Group 4 Fax
LZW = 5  # Lempel-Ziv-Welch
JPEG_OLD = 6  # old-style JPEG
JPEG = 7
ADOBE_DEFLATE = 8
JBIG_BW = 9  # VC5
# JBIG_COLOR = 10
# JPEG_99 = 99
# KODAK_262 = 262
JPEGXR_NDPI = 22610  # JPEG XR (Hammamatsu NDPI)
# NEXT = 32766
# SONY_ARW = 32767
# PACKED_RAW = 32769
CCIRLEW = 32771  # Word-aligned 1D Huffman compression
PACKBITS = 32773  # PackBits, aka Macintosh RLE
THUNDERSCAN = 32809
# IT8LW = 32896
# IT8MP = 32897
# IT8BL = 32898
# PIXARFILM = 32908
# PIXARLOG = 32909
DEFLATE = 32946
DCS = 32947
JPEG2000_YUV = 33003  # JPEG 2000 YCbCr (Leica Aperio)
JPEG2000_LOSSY = 33004  # Lossy JPEG 2000 (Bio-Formats)
JPEG2000_RGB = 33005  # JPEG 2000 RGB (Leica Aperio)
ALT_JPEG = 33007  # JPEG (Bio-Formats)
JBIG = 34661
SGILOG = 34676  # LogLuv32
SGILOG24 = 34677
JPEG2000 = 34712
# NIKON_NEF = 32713
# JBIG2 = 32715
# MDI_BINARY = 34718  # Microsoft Document Imaging
# MDI_PROGRESSIVE = 34719  # Microsoft Document Imaging
# MDI_VECTOR = 34720  # Microsoft Document Imaging
LERC = 34887  # ESRI Limited Error Raster Compression
JPEG_LOSSY = 34892  # DNG
LZMA = 34925  # Lempel-Ziv-Markov chain Algorithm
ZSTD_DEPRECATED = 34926  # deprecated
WEBP_DEPRECATED = 34927  # deprecated
PNG = 34933  # Portable Network Graphics (Zoomable Image File format)
JPEGXR = 34934  # JPEG XR (Zoomable Image File format)
JETRAW = 48124  # Jetraw by Dotphoton
ZSTD = 50000  # Zstandard
WEBP = 50001  # WebP
JPEGXL = 50002  # GDAL
PIXTIFF = 50013  # ZLIB (Atalasoft)
JPEGXL_DNG = 52546  # JPEG XL (DNG)
EER_V0 = 65000  # FIXED82 Thermo Fisher Scientific
EER_V1 = 65001  # FIXED72 Thermo Fisher Scientific
EER_V2 = 65002  # VARIABLE Thermo Fisher Scientific
# PENTAX_PEF = 65535



from ._mosaic import Mosaic
from ._norm import LumaScaler, Normalizer
from ._util import (
    at,
    get_fusion,
    get_trapz,
    normalize_loc,
    probs_to_rgb_heatmap,
    rescale_crop,
    resize,
)

__all__ = [
    'LumaScaler',
    'Mosaic',
    'Normalizer',
    'at',
    'get_fusion',
    'get_trapz',
    'normalize_loc',
    'probs_to_rgb_heatmap',
    'rescale_crop',
    'resize',
]



__all__ = ['Mosaic', 'Patches', 'Tiles']

import dataclasses
import warnings
from collections.abc import Callable, Iterable, Iterator, Sequence
from contextlib import AbstractContextManager
from dataclasses import dataclass
from functools import partial
from itertools import chain, starmap
from math import ceil
from typing import Literal, Self, cast

import cv2
import numpy as np
import numpy.typing as npt
from glow import chunked, map_n, starmap_n

from bipl._types import HasPartsAbc, NDIndex, NumpyLike, Shape, Span, Tile, Vec

from ._tile import BlendCropper, Decimator, Reconstructor, Stripper, Zipper
from ._util import at, get_trapz, padslice

# TODO: allow result of .map/.map_batched to have different tile and step
# NOTE: all classes here are stateless & their iterators are reentrant.

# ---------------------------- utility functions -----------------------------


def _apply(fn: Callable[[np.ndarray], np.ndarray], obj: Tile) -> Tile:
    r = fn(obj.data)
    assert r.shape[:2] == obj.data.shape[:2], 'Callable changed data shape'
    return obj._replace(data=r)


def _apply_batched(
    fn: Callable[[list[np.ndarray]], Iterable[np.ndarray]],
    ts: tuple[Tile, ...],
) -> list[Tile]:
    [*rs] = fn([t.data for t in ts])
    assert len(rs) == len(ts), 'Input & output has different sizes'

    r_shapes = [r.shape[:2] for r in rs]
    t_shapes = [t.data.shape[:2] for t in ts]
    assert r_shapes == t_shapes, 'Tile height/width should not change'

    return [t._replace(data=r) for t, r in zip(ts, rs)]


def _reweight(weight: np.ndarray, tile: np.ndarray) -> np.ndarray:
    assert tile.dtype.kind == 'f'
    return np.einsum('hwc,h,w -> hwc', tile, weight, weight, optimize=True)


# ------------------------------ mosaic setting ------------------------------


@dataclass(frozen=True, slots=True)
class Mosaic:
    """
    Helper to split image to tiles and process them.

    Parameters:
    - step - Step between consecutive tiles
    - overlap - Count of pixels that will be shared among overlapping tiles

    So tile size = overlap + non-overlapped area + overlap = step + overlap,
    and non-overlapped area = step - overlap.
    """

    step: int
    overlap: int

    def __post_init__(self):
        if self.overlap < 0:
            raise ValueError(
                f'overlap should be non-negative, got: {self.overlap}'
            )
        # That may be optional
        if self.overlap % 2 != 0:
            raise ValueError(f'overlap should be even, got: {self.overlap}')
        if self.overlap > self.step:
            raise ValueError(
                'overlap should be lower than step, got: '
                f'overlap={self.overlap} and step={self.step}'
            )

    @property
    def tile_size(self) -> int:
        return self.step + self.overlap

    def get_kernel(self) -> np.ndarray:
        return get_trapz(self.step, self.overlap)

    def iterate(self, image: NumpyLike, max_workers: int = 1) -> '_ArrayTiles':
        """Read tiles from input image"""
        # Source shape
        shape = tuple(image.shape[:2])

        # Index
        ih, iw = ishape = tuple(
            (s + self.tile_size - 1) // self.step for s in shape
        )
        cells = np.ones(ishape, dtype=np.bool_)

        # Align slide & cells centers
        view = tuple(round(i * self.step + self.overlap) for i in ishape)
        origin = tuple((s - v) // 2 for s, v in zip(shape, view))

        # All of (ih iw YX)
        # First tile is [origin: origin + tile]
        iyx = np.mgrid[:ih, :iw].transpose(1, 2, 0)
        yx0 = self.step * iyx + origin
        yx1 = self.tile_size + yx0
        locs = np.stack([yx0, yx1], -1)  # (ih iw YX lo-hi)

        return _ArrayTiles(shape, self, cells, locs, image, max_workers)


# ---------------------------------- pathes ----------------------------------


@dataclass(frozen=True)
class Patches:
    """Iterable of rectangular patches from non-uniform grid"""

    shape: Shape

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> Iterator[Tile]:
        raise NotImplementedError

    def _patches_from(self, it: Iterable[Tile]) -> 'Patches':
        return _IterPatches(self.shape, it, len(self))

    def _new_from(self, it: Iterable[Tile]) -> Self:
        return cast(Self, self._patches_from(it))

    def map(
        self, fn: Callable[[np.ndarray], np.ndarray], /, max_workers: int = 0
    ) -> Self:
        """
        Apply function to each patch/tile.
        Note: height/width of patch must not be changed.
        Each patch is HWC-ordered ndarray.
        Supports threading.
        """
        return self._new_from(
            map_n(partial(_apply, fn), self, max_workers=max_workers)
        )

    def pool(self, stride: int = 1) -> 'Patches':
        """Resize each patch to desired stride"""
        if stride == 1:
            return self
        shape = tuple((s + stride - 1) // stride for s in self.shape)
        return _IterPatches(shape, _Decimated(self, stride), len(self))

    def transform(
        self,
        fn: Callable[[Iterable[Tile]], Iterable[Tile]],
    ) -> 'Patches':
        """
        TODO: fill docs
        """
        return self._patches_from(fn(self))

    def strip(self) -> 'Patches':
        """Strip out-of-bounds regions. Produces non-square patches"""
        return self._patches_from(map(Stripper(self.shape), self))

    def crop(self) -> 'Patches':
        warnings.warn(
            '"Patches.crop" is deprecated. Use "Patches.strip"',
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.strip()

    def zip_with(
        self,
        view: NumpyLike,
        v_scale: float,
        interpolation: Literal[0, 1, 2, 3, 4] = 0,
    ) -> '_ZipView':
        """
        For each tile read & resample matching region from `view`.
        Similar to RoIAlign used in some detectors.

        Interpolation is Nearest (0) by default, but Linear (1), Bicubic (2),
        Area (3) & Lanczos-4 (4) are also supported (OpenCV codes).
        """
        if v_scale <= 0:
            raise ValueError(f'v_scale should be positive, got: {v_scale}')
        return _ZipView(self, len(self), view, v_scale, interpolation)

    def with_(self, ctx: AbstractContextManager) -> Self:
        """Iterate with context. Reentrant if nested context is reentrant"""
        return self._new_from(_Scoped(self, ctx))


@dataclass(frozen=True)
class _ZipView:
    source: Iterable[Tile]
    total: int
    view: NumpyLike
    v_scale: float
    interpolation: int

    def __len__(self) -> int:
        return self.total

    def __iter__(
        self,
    ) -> Iterator[tuple[NDIndex, Vec, np.ndarray, np.ndarray]]:
        return map(
            Zipper(self.view, self.v_scale, self.interpolation),
            self.source,
        )


@dataclass(frozen=True)
class _IterPatches(Patches):
    source: Iterable[Tile]
    total: int

    def __len__(self) -> int:
        return self.total

    def __iter__(self) -> Iterator[Tile]:
        return iter(self.source)


# ---------------------------------- tiles -----------------------------------


@dataclass(frozen=True)
class Tiles(Patches):
    """Iterable of (overlapping) square tiles from uniform grid"""

    m: Mosaic
    cells: npt.NDArray[np.bool_]

    @property
    def ishape(self) -> Sequence[int]:
        return self.cells.shape

    def __len__(self) -> int:
        return int(self.cells.sum())

    def report(self) -> dict[str, str]:
        """Cells and area usage"""
        used = int(self.cells.sum())
        total = self.cells.size
        coverage = (used / total) * (1 + self.m.overlap / self.m.step) ** 2
        return {'cells': f'{used}/{total}', 'coverage': f'{coverage:.0%}'}

    def _new_from(self, it: Iterable[Tile]) -> 'Tiles':
        return _IterTiles(self.shape, self.m, self.cells, it)

    def map_batched(
        self,
        fn: Callable[[list[np.ndarray]], Iterable[np.ndarray]],
        /,
        batch_size: int = 1,
        max_workers: int = 0,
    ) -> 'Tiles':
        """
        Apply function to batches of tiles.
        Note: change of tile shape besides channel count is forbidden.
        Each tile is HWC-ordered ndarray.
        Supports threading.
        """
        tile_fn = partial(_apply_batched, fn)

        chunks = chunked(self, batch_size)
        batches = map_n(tile_fn, chunks, max_workers=max_workers)
        tiles = chain.from_iterable(batches)

        return self._new_from(tiles)

    def pool(self, stride: int = 1) -> 'Tiles':
        """Resize each tile to desired stride"""
        if stride == 1:
            return self
        if self.m.step % stride != 0 or self.m.overlap % stride != 0:
            raise ValueError(
                'stride should be a divisor of '
                f'{self.m.step}, {self.m.overlap}. Got: {stride}'
            )

        m = Mosaic(self.m.step // stride, self.m.overlap // stride)
        shape = tuple((s + stride - 1) // stride for s in self.shape)
        return _IterTiles(shape, m, self.cells, _Decimated(self, stride))

    def reweight(self) -> 'Tiles':
        """
        Apply weight to tile edges to prepare them for summation
        if overlap exists.
        Note: no need to call this method if you have already applied it.
        """
        if not self.m.overlap:  # No-op
            return self
        weight = self.m.get_kernel()
        tile_fn = partial(_reweight, weight)
        return self.map(tile_fn)

    def merge(self) -> Patches:
        """
        Remove overlapping regions from all the tiles if any.
        Tiles should be reweighted if overlap exists.
        """
        if not self.m.overlap:  # No-op
            return self
        bcr = BlendCropper(self.cells, self.m.overlap, self.m.step)
        return _IterPatches(self.shape, _Merged(self, bcr), bcr.size)


@dataclass(frozen=True)
class _IterTiles(Tiles):
    source: Iterable[Tile]

    def __iter__(self) -> Iterator[Tile]:
        return iter(self.source)


@dataclass(frozen=True, slots=True)
class _Scoped:
    source: Iterable[Tile]
    _ctx: AbstractContextManager

    def __iter__(self) -> Iterator[Tile]:
        with self._ctx:
            yield from self.source


@dataclass(frozen=True, slots=True)
class _Decimated:
    """
    Decimates tiles.
    Doesn't change size uniformity.
    Yields decimated views of original tiles.
    """

    source: Iterable[Tile]
    stride: int

    def __iter__(self) -> Iterator[Tile]:
        return map(Decimator(self.stride), self.source)


@dataclass(frozen=True, slots=True)
class _Merged:
    """
    Sums overlapping regions.
    Returns non-uniform patches without overlaps.
    """

    source: Iterable[Tile]
    bcr: BlendCropper

    def __iter__(self) -> Iterator[Tile]:
        return self.bcr(self.source)


@dataclass(frozen=True)
class _ArrayTiles(Tiles):
    """Extracts square tiles from array with overlaps."""

    locs: npt.NDArray[np.int64]  # (ih iw YX lo-hi)
    data: NumpyLike
    max_workers: int

    def select(
        self, mask: np.ndarray, scale: float | None = None
    ) -> '_ArrayTiles':
        """
        Subset non-masked (i.e. non-zeros in mask) tiles for iteration.
        Mask size is "scale"-multiple of source image.

        NOTE: `select(mask1).select(mask2)` is equal to `select(mask1 & mask2)`
        """
        if mask.ndim == 3:  # Strip extra dim
            mask = mask.squeeze(2)
        if mask.ndim != 2:
            raise ValueError(f'Mask should be 2D, got shape: {mask.shape}')
        if scale is None:
            scale = max(ms / s for ms, s in zip(mask.shape, self.data.shape))

        mask = np.where(mask, np.uint8(255), np.uint8(0))

        if half_tile := ceil(self.m.tile_size * scale / 2):
            kernel = np.ones((3, 3), dtype='u1')
            mask = cv2.dilate(mask, kernel, iterations=half_tile)

        # Align mask & cells centers
        ih, iw = ishape = self.cells.shape
        view = tuple(round(i * self.m.step * scale) for i in ishape)
        loc = tuple(
            ((s0 - s1) // 2, (s0 - s1) // 2 + s1)
            for s0, s1 in zip(mask.shape, view)
        )
        mask = padslice(mask, *loc)

        cells = cv2.resize(mask, (iw, ih), interpolation=cv2.INTER_AREA)

        return dataclasses.replace(self, cells=self.cells & cells.astype(bool))

    def _get_tile(self, idx: NDIndex, *loc: Span) -> Tile:
        """Do `data[loc]`. Result could be cropped."""
        vec = tuple(lo for lo, _ in loc)
        return Tile(idx=idx, vec=vec, data=at(self.data, *loc))

    def _get_tile_padded(self, idx: NDIndex, *loc: Span) -> Tile:
        """Do `data[loc]` padding data in necessary."""
        vec = tuple(lo for lo, _ in loc)
        return Tile(idx=idx, vec=vec, data=padslice(self.data, *loc))

    def _ilocs(
        self,
        locs: npt.NDArray[np.int64],
    ) -> list[tuple[NDIndex, Span, Span]]:
        """Precompute tile coordinates: 2D index & Y/X-slices"""
        iys, ixs = self.cells.nonzero()
        boxes = locs[iys, ixs].tolist()
        return [
            (i, tuple(ys), tuple(xs))
            for i, (ys, xs) in zip(zip(iys.tolist(), ixs.tolist()), boxes)
        ]

    def _drop_overlaps(
        self, grid: npt.NDArray[np.int64]
    ) -> npt.NDArray[np.int64]:
        grid = grid.copy()  # (ih iw YX lo-hi)

        # Cut top edge (y0) from tiles having neighbor from above (iy)
        grid[1:, :, 0, 0][self.cells[:-1, :]] += self.m.overlap

        # Cut left edge (x0) from tiles having neighbor on the left (ix)
        grid[:, 1:, 1, 0][self.cells[:, :-1]] += self.m.overlap
        return grid

    def __iter__(self) -> Iterator[Tile]:
        """
        Yield complete tiles built from source image.
        Each tile will have size `(step + overlap)`
        """
        if isinstance(self.data, np.ndarray):
            ilocs = self._ilocs(self.locs)
            return starmap(self._get_tile_padded, ilocs)

        # De-overlap if we don't know whether `data` has cheap slices
        # Though we hope that data don't use index wrapping i.e. `index % size`
        locs = self._drop_overlaps(self.locs)
        ilocs = self._ilocs(locs)
        if isinstance(self.data, HasPartsAbc):
            ids = [i[0] for i in ilocs]
            boxes = [i[1:] for i in ilocs]
            patches = self.data.parts(boxes, max_workers=self.max_workers)
            parts = (
                Tile(idx=i, vec=tuple(s[0] for s in p.loc), data=p.data)
                for i, p in zip(ids, patches)
            )
        else:
            parts = starmap_n(
                self._get_tile, ilocs, max_workers=self.max_workers
            )

        if not self.m.overlap:
            return parts

        rcr = Reconstructor(self.m.overlap, self.cells)
        return map(rcr, parts)

    def get_tile(self, idx: int) -> Tile:
        """Get full tile by its flattened index"""
        iys, ixs = self.cells.nonzero()
        i = int(iys[idx]), int(ixs[idx])
        ys, xs = self.locs[i].tolist()
        return self._get_tile_padded(i, ys, xs)



__all__ = [
    'at',
    'crop_to',
    'get_fusion',
    'get_trapz',
    'merge_intervals',
    'normalize_loc',
    'probs_to_rgb_heatmap',
    'resize',
]

from collections.abc import Iterable, Sequence
from itertools import starmap
from math import ceil, floor

import cv2
import numpy as np

from bipl import env
from bipl._types import NumpyLike, Span, Tile, Vec


def probs_to_rgb_heatmap(prob: np.ndarray) -> np.ndarray:
    """Converts probability array of (H W C) shape to (H W 3) RGB image"""
    if prob.ndim != 3:
        raise ValueError(f'prob should be 3d, got: {prob.shape}')
    h, w, c = prob.shape
    if not prob.size:
        return np.empty((h, w, 3), dtype='u1')

    hue = prob.argmax(-1).astype('u1')  # NOTE: valid only for simple Indexer
    value = prob.max(-1)
    if value.dtype != 'u1':
        value *= 255
        value = value.round().astype('u1')

    hue_lut = np.zeros(256, 'u1')
    hue_lut[:c] = np.linspace(0, 127, c, endpoint=False, dtype='u1')
    hsv = cv2.merge(
        [
            cv2.LUT(hue, hue_lut),
            np.broadcast_to(np.uint8(255), (h, w)),
            value,
        ]
    )
    return cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)


def get_trapz(step: int, overlap: int) -> np.ndarray:
    """Returns trapezional window kernel to apply stiching"""
    if overlap == 0:
        raise ValueError('not applicable for overlap 0')

    ramp = np.arange(1, overlap + 1).astype('f4')
    ramp /= overlap + 1

    r = np.empty(step + overlap, dtype='f4')
    r[:overlap] = ramp
    r[overlap:step] = 1
    r[step:] = ramp[::-1]
    return r


def normalize_loc(
    loc: Sequence[slice] | slice, shape: Sequence[int]
) -> tuple[Span, ...]:
    """Ensures slices match ndim and have noo `None` endpoints"""
    if isinstance(loc, slice):
        loc = (loc,)
    if any(s.step not in (None, 1) for s in loc):
        raise ValueError(f'Slice step is not supported, got: {loc}')
    loc_ = [
        (s.start or 0, size if s.stop is None else s.stop)
        for s, size in zip(loc, shape)
    ]
    return _pad_loc(loc_, shape)


def _pad_loc(loc: Sequence[Span], shape: Sequence[int]) -> tuple[Span, ...]:
    n = len(loc)
    ndim = len(shape)
    if n > ndim:
        raise ValueError(f'loc is too deep for {ndim}D, got: {loc}')
    return *loc, *((0, s) for s in shape[n:])


def padslice(a: NumpyLike, *loc: Span) -> np.ndarray:
    """
    Do `a[loc]`, but extend `a` (with 0s) if `loc` indices beyond `a.shape`.
    """
    loc = _pad_loc(loc, a.shape)
    r_shape = [hi - lo for lo, hi in loc]

    # Nothing to pad
    if not all(a.shape):
        return np.zeros(r_shape, a.dtype)

    # No need to pad
    if all(0 <= lo <= hi <= size for (lo, hi), size in zip(loc, a.shape)):
        return at(a, *loc)

    iloc = tuple(slice(*np.clip(s, 0, size)) for s, size in zip(loc, a.shape))
    rloc = tuple(
        slice(i.start - lo, i.stop - lo) for (lo, _), i in zip(loc, iloc)
    )

    r = np.empty(r_shape, a.dtype)
    pad: tuple[slice, ...] = ()
    for o in rloc:
        r[*pad, : o.start] = 0
        r[*pad, o.stop :] = 0
        pad += (o,)
    r[rloc] = a[iloc]
    return r


def crop_to(
    vec: Vec, a: NumpyLike, shape: Sequence[int]
) -> tuple[Vec, np.ndarray]:
    """
    Crop `a` to be completely within shape, i.e.
    `0 <= vec[i] <= vec[i] + a.shape[i] <= shape[i]`.

    Arguments:
    - vec - top-left offset of `a` w.r.t. target.
    - a - data to crop.
    - shape - target shape to fit into.

    Returns new offset & data.
    """
    assert len(a.shape) >= len(shape), 'Desired shape has more dims than `a`'
    loc = tuple(
        tuple(np.clip([0, ts], -t0, s - t0))
        for t0, ts, s in zip(vec, a.shape, shape)
    )
    a = padslice(a, *loc)

    vec = tuple(max(v, 0) for v in vec)
    return vec, a


def resize(
    image: np.ndarray, hw: Sequence[int], *, antialias: bool | None = None
) -> np.ndarray:
    if image.shape[:2] == hw[:2]:
        return image

    f_max = max(s0 / s1 for s0, s1 in zip(image.shape, hw))
    if f_max < 1:  # Upscaling
        antialias = False
    elif antialias is None:  # Enable only if we downscale more than 2.5x
        antialias = f_max > 2.5

    h, w = hw
    interpolation = cv2.INTER_LINEAR
    if antialias:  # Downsampling
        match env.BIPL_DOWN:
            case 'box':  # Pyramid, box filter. Fastest
                for _ in range(int(f_max).bit_length() - 1):
                    image = cv2.resize(image, None, fx=0.5, fy=0.5)

            case 'gauss':  # Pyramid, gauss filter. No aliasing. 2X slower
                for _ in range(int(f_max).bit_length() - 1):
                    image = cv2.pyrDown(image)

            case 'area':  # No pyramid, box filter. Slowest
                interpolation = cv2.INTER_AREA

    return cv2.resize(image, (w, h), interpolation=interpolation)


def get_fusion(
    tiles: Iterable[Tile], shape: Sequence[int] | None = None
) -> np.ndarray | None:
    """Stack tiles to large image. Tiles lying outside of bounds are cropped"""
    r: np.ndarray | None = None

    if shape is None:  # Collect all the tiles to compute destination size
        tiles = [*tiles]
        if not tiles:
            return None
        # N arrays of (yx + hw)
        yx_hw = np.array([[[t.vec, t.data.shape[:2]] for t in tiles]]).sum(1)
        # bottom left most edge
        shape = tuple(yx_hw.max(0).tolist())
    elif len(shape) != 2:
        raise ValueError(f'shape should be 2-tuple, got {shape}')

    for _, vec, tile in tiles:
        (y, x), a = crop_to(vec, tile, shape)
        if not a.size:
            continue

        h, w, c = a.shape
        if r is None:  # First iteration, initialize
            r = np.zeros((*shape, c), a.dtype)
        if c != r.shape[2]:
            raise RuntimeError('tile channel counts changed during iteration')
        r[y : y + h, x : x + w] = a

    return r


def rescale_crop(
    a: NumpyLike, *loc: Span, scale: float = 1, interpolation: int = 0
):
    """
    Rescale image, then crop with respect to subpixels.

    Interpolation is Nearest (0) by default, but Linear (1), Bicubic (2),
    Area (3) & Lanczos-4 (4) are also supported (OpenCV codes).

    Effectively is same as:
    ```
    (y0, y1), (x0, x1) = ys * scale, xs * scale
    h, w = ys[1] - ys[0], xs[1] - xs[0]
    return cv2.resize(a[y0: y1, x0: x1], (w, h), interpolation=interpolation)
    ```
    """
    assert 0 <= interpolation <= 4
    if scale == 1:
        return padslice(a, *loc)

    box = np.array(loc, 'i4')  # (y/x, start/stop)
    h, w = (box[:, 1] - box[:, 0]).tolist()
    if not h or not w:  # 0-size output
        return np.empty((h, w, *a.shape[2:]), a.dtype)

    lo_f, hi_f = np.multiply(box, scale, dtype='f4').T

    loc = tuple((floor(lo_), ceil(hi_)) for lo_, hi_ in zip(lo_f, hi_f))
    if not env.BIPL_SUBPIX:
        return resize(padslice(a, *loc), (h, w))

    # Extra margin to accomodate interpolation kernel
    # 2x2 (0 extra) - nearest (0), bilinear (1), area (3)
    # 4x4 (1 extra) - bicubic (2)
    # 8x8 (3 extra) - lanczos4 (4)
    eps = (0, 0, 1, 0, 3)[interpolation]

    # Tight slice to have all necessary pixels
    loc = tuple(
        tuple(np.clip([lo - eps, hi + eps], 0, size))
        for (lo, hi), size in zip(loc, a.shape)
    )
    r = at(a, *loc)
    if not r.size:  # 0-size tight slice
        return np.zeros((h, w, *a.shape[2:]), a.dtype)

    # Resample image crop to destination grid
    dy, dx = (lo_ - lo + (scale - 1) / 2 for lo_, (lo, _) in zip(lo_f, loc))
    return cv2.warpAffine(
        r,
        np.array([[scale, 0, dx], [0, scale, dy]], 'f4'),
        (w, h),
        flags=interpolation | cv2.WARP_INVERSE_MAP,
        borderMode=cv2.BORDER_CONSTANT,
    )


def at(a: NumpyLike, *loc: Span) -> np.ndarray:
    s = tuple(starmap(slice, loc))
    return a[s]


def merge_intervals(spans: list[Span]) -> tuple[list[int], list[Span]]:
    n = len(spans)
    if n == 1:
        return [*range(n)], spans

    # Ascening order of span.start
    spans_a = np.asarray(spans)
    pos = spans_a[:, 0].argsort()  # (n)
    spans_a = spans_a[pos]  # (n 2)

    # "Merge intervals" algorithm for non-overlapping intervals
    # See https://stackoverflow.com/a/58976449/9868257
    edges = spans_a[:-1, 1] < spans_a[1:, 0]  # (n-1)
    edges = np.r_[True, edges, True]  # (n+1)
    mask = np.ndarray((n, 2), '?', buffer=edges, strides=edges.strides * 2)
    spans_a = spans_a[mask].reshape(-1, 2)

    return pos.tolist(), spans_a.tolist()
